from __future__ import annotations

import datetime
import json
from typing import Dict, Optional, List, Any, TYPE_CHECKING, Callable, Set
import requests

from ravendb.primitives.constants import int_max
from ravendb.documents.session.loaders.include import TimeSeriesIncludeBuilder
from ravendb.documents.session.time_series import TimeSeriesEntry, AbstractTimeSeriesRange
from ravendb.http.http_cache import HttpCache
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import RaftCommand
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.documents.operations.definitions import MaintenanceOperation, IOperation, VoidOperation
from ravendb.primitives.time_series import TimeValue
from ravendb.tools.utils import Utils, CaseInsensitiveDict
from ravendb.util.util import RaftIdGenerator
from ravendb.documents.conventions import DocumentConventions

if TYPE_CHECKING:
    from ravendb.documents.store.definition import DocumentStore


class TimeSeriesPolicy:
    def __init__(
        self,
        name: Optional[str] = None,
        aggregation_time: Optional[TimeValue] = None,
        retention_time: TimeValue = TimeValue.MAX_VALUE(),
    ):
        if not name or name.isspace():
            raise ValueError("Name cannot be None or empty")

        if aggregation_time and aggregation_time.compare_to(TimeValue.ZERO()) <= 0:
            raise ValueError("Aggregation time must be greater than zero")

        if retention_time is None or retention_time.compare_to(TimeValue.ZERO()) <= 0:
            raise ValueError("Retention time must be greater than zero")

        self.retention_time = retention_time
        self.aggregation_time = aggregation_time

        self.name = name

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesPolicy:
        return cls(
            json_dict["Name"],
            TimeValue.from_json(json_dict["AggregationTime"]),
            TimeValue.from_json(json_dict["RetentionTime"]),
        )

    def get_time_series_name(self, raw_name: str) -> str:
        return raw_name + TimeSeriesConfiguration.TIME_SERIES_ROLLUP_SEPARATOR + self.name

    def to_json(self) -> Dict[str, Any]:
        return {
            "Name": self.name,
            "AggregationTime": self.aggregation_time.to_json() if self.aggregation_time else None,
            "RetentionTime": self.retention_time.to_json(),
        }


class RawTimeSeriesPolicy(TimeSeriesPolicy):
    POLICY_STRING = "rawpolicy"  # must be lower case

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> RawTimeSeriesPolicy:
        return cls(TimeValue.from_json(json_dict["RetentionTime"]))

    @classmethod
    def DEFAULT_POLICY(cls) -> RawTimeSeriesPolicy:
        return cls(TimeValue.MAX_VALUE())

    def __init__(self, retention_time: TimeValue = TimeValue.MAX_VALUE()):
        if retention_time.compare_to(TimeValue.ZERO()) <= 0:
            raise ValueError("Retention time must be greater than zero")
        super().__init__(self.POLICY_STRING, retention_time=retention_time)


class TimeSeriesCollectionConfiguration:
    def __init__(
        self,
        disabled: Optional[bool] = False,
        policies: Optional[List[TimeSeriesPolicy]] = None,
        raw_policy: Optional[RawTimeSeriesPolicy] = RawTimeSeriesPolicy.DEFAULT_POLICY(),
    ):
        self.disabled = disabled
        self.policies = policies
        self.raw_policy = raw_policy

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesCollectionConfiguration:
        return cls(
            json_dict["Disabled"],
            [TimeSeriesPolicy.from_json(policy_json) for policy_json in json_dict["Policies"]],
            RawTimeSeriesPolicy.from_json(json_dict["RawPolicy"]),
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "Disabled": self.disabled,
            "Policies": [policy.to_json() for policy in self.policies],
            "RawPolicy": self.raw_policy.to_json(),
        }


class TimeSeriesConfiguration:
    TIME_SERIES_ROLLUP_SEPARATOR = "@"

    @classmethod
    def from_json(
        cls,
        json_dict: Dict[str, Any] = None,
    ) -> TimeSeriesConfiguration:
        configuration = cls()
        configuration.collections = {
            key: TimeSeriesCollectionConfiguration.from_json(value) for key, value in json_dict["Collections"].items()
        }
        configuration.policy_check_frequency = (
            Utils.string_to_timedelta(json_dict["PolicyCheckFrequency"])
            if "PolicyCheckFrequency" in json_dict and json_dict["PolicyCheckFrequency"]
            else None
        )
        configuration.named_values = json_dict["NamedValues"]
        configuration._internal_post_json_deserialization()
        return configuration

    def __init__(self):
        self.collections: Dict[str, TimeSeriesCollectionConfiguration] = {}
        self.policy_check_frequency: Optional[datetime.timedelta] = None
        self.named_values: Optional[Dict[str, Dict[str, List[str]]]] = None

    def to_json(self) -> Dict[str, Any]:
        return {
            "Collections": {key: value.to_json() for key, value in self.collections.items()},
            "PolicyCheckFrequency": Utils.timedelta_to_str(self.policy_check_frequency),
            "NamedValues": self.named_values,
        }

    def get_names(self, collection: str, time_series: str) -> Optional[List[str]]:
        if self.named_values is None:
            return None

        ts_holder = self.named_values.get(collection, None)
        if ts_holder is None:
            return None

        names = ts_holder.get(time_series, None)
        if names is None:
            return None

        return names

    def _internal_post_json_deserialization(self) -> None:
        self._populate_named_values()
        self._populate_policies()

    def _populate_policies(self) -> None:
        if self.collections is None:
            return

        dic = CaseInsensitiveDict()
        for key, value in self.collections.items():
            dic[key] = value

        self.collections = dic

    def _populate_named_values(self) -> None:
        if self.named_values is None:
            return

        # ensure ignore case
        dic = CaseInsensitiveDict()

        for key, value in self.named_values.items():
            value_map = CaseInsensitiveDict()
            value_map.update(value)
            dic[key] = value_map

        self.named_values = dic


class ConfigureTimeSeriesOperationResult:
    def __init__(self, raft_command_index: int):
        self.raft_command_index = raft_command_index

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]):
        return cls(json_dict["RaftCommandIndex"])


class ConfigureTimeSeriesPolicyOperation(MaintenanceOperation[ConfigureTimeSeriesOperationResult]):
    def __init__(self, collection: str, config: TimeSeriesPolicy):
        self.collection = collection
        self.config = config

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[ConfigureTimeSeriesOperationResult]":
        return self.ConfigureTimeSeriesPolicyCommand(self.collection, self.config)

    class ConfigureTimeSeriesPolicyCommand(RavenCommand[ConfigureTimeSeriesOperationResult], RaftCommand):
        def __init__(self, collection: str, configuration: TimeSeriesPolicy):
            if not configuration:
                raise ValueError("Configuration cannot be None")
            if not collection or collection.isspace():
                raise ValueError("Collection cannot be None or empty")

            super().__init__(ConfigureTimeSeriesOperationResult)

            self._collection = collection
            self._configuration = configuration

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            request = requests.Request(
                "PUT",
                f"{node.url}/databases/{node.database}/admin/timeseries" f"/policy?collection={self._collection}",
                data=self._configuration.to_json(),
            )
            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()
            self.result = ConfigureTimeSeriesOperationResult.from_json(json.loads(response))

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class ConfigureRawTimeSeriesPolicyOperation(ConfigureTimeSeriesPolicyOperation):
    def __init__(self, collection: str, config: RawTimeSeriesPolicy):
        super(ConfigureRawTimeSeriesPolicyOperation, self).__init__(collection, config)


class ConfigureTimeSeriesOperationResult:
    def __init__(self, raft_command_index: int = None):
        self.raft_command_index = raft_command_index

    def to_json(self) -> Dict:
        return {
            "RaftCommandIndex": self.raft_command_index,
        }

    @classmethod
    def from_json(cls, json_dict: Dict) -> ConfigureTimeSeriesOperationResult:
        return cls(json_dict.get("RaftCommandIndex"))


class ConfigureTimeSeriesValueNamesOperation(MaintenanceOperation[ConfigureTimeSeriesOperationResult]):
    class Parameters:
        def __init__(self, collection: str, time_series: str, value_names: List[str], update: Optional[bool] = None):
            self.collection = collection
            self.time_series = time_series
            self.value_names = value_names
            self.update = update

        def validate(self) -> None:
            if not self.collection or self.collection.isspace():
                raise ValueError("Collection name cannot be None or empty")
            if not self.time_series or self.time_series.isspace():
                raise ValueError("TimeSeries cannot be None or empty")
            if not self.value_names:
                raise ValueError("ValueNames cannot be None or empty")

        def to_json(self) -> Dict[str, Any]:
            return {
                "Collection": self.collection,
                "TimeSeries": self.time_series,
                "ValueNames": self.value_names,
                "Update": self.update,
            }

    def __init__(self, parameters: Parameters):
        if parameters is None:
            raise ValueError("Parameters cannot be None")
        super(ConfigureTimeSeriesValueNamesOperation, self).__init__()
        self._parameters = parameters
        self._parameters.validate()

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[ConfigureTimeSeriesOperationResult]":
        return self.ConfigureTimeSeriesValueNamesCommand(self._parameters)

    class ConfigureTimeSeriesValueNamesCommand(RavenCommand[ConfigureTimeSeriesOperationResult], RaftCommand):
        def __init__(self, parameters: ConfigureTimeSeriesValueNamesOperation.Parameters):
            super().__init__(ConfigureTimeSeriesOperationResult)
            self._parameters = parameters

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/timeseries/names/config"
            request = requests.Request("POST", url)
            request.data = self._parameters.to_json()
            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if not response:
                self._throw_invalid_response()

            self.result = ConfigureTimeSeriesOperationResult.from_json(json.loads(response))

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class RemoveTimeSeriesPolicyOperation(MaintenanceOperation[ConfigureTimeSeriesOperationResult]):
    def __init__(self, collection: str, name: str):
        if not collection or collection.isspace():
            raise ValueError("Collection cannot be None or empty")

        if not name or name.isspace():
            raise ValueError("Name cannot be None or empty")

        self._collection = collection
        self._name = name

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[ConfigureTimeSeriesOperationResult]:
        return self.RemoveTimeSeriesPolicyCommand(self._collection, self._name)

    class RemoveTimeSeriesPolicyCommand(RavenCommand[ConfigureTimeSeriesOperationResult], RaftCommand):
        def __init__(self, collection: str, name: str):
            super().__init__(ConfigureTimeSeriesOperationResult)
            self._collection = collection
            self._name = name

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "DELETE",
                f"{node.url}/databases"
                f"/{node.database}/admin/timeseries/policy?"
                f"collection={self._collection}&"
                f"name={self._name}",
            )

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = ConfigureTimeSeriesOperationResult.from_json(json.loads(response))

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class TimeSeriesOperation:
    class AppendOperation:
        def __init__(self, timestamp: datetime.datetime, values: List[float], tag: Optional[str] = None):
            self.timestamp = timestamp
            self.values = values
            self.tag = tag

        def __eq__(self, other):
            return isinstance(other, TimeSeriesOperation.AppendOperation) and other.timestamp == self.timestamp

        def __hash__(self):
            return hash(self.timestamp)

        def to_json(self) -> Dict[str, Any]:
            json_dict = {
                "Timestamp": Utils.datetime_to_string(self.timestamp),
                "Values": self.values,
            }
            if self.tag:
                json_dict.update({"Tag": self.tag})

            return json_dict

    class DeleteOperation:
        def __init__(
            self, datetime_from: Optional[datetime.datetime] = None, datetime_to: Optional[datetime.datetime] = None
        ):
            self._datetime_from = datetime_from
            self._datetime_to = datetime_to

        def to_json(self) -> Dict[str, Any]:
            return {
                "From": Utils.datetime_to_string(self._datetime_from) if self._datetime_from else None,
                "To": Utils.datetime_to_string(self._datetime_to) if self._datetime_to else None,
            }

    def __init__(self, name: Optional[str] = None):
        self.name = name
        self._appends: Set[TimeSeriesOperation.AppendOperation] = set()
        self._deletes: List[TimeSeriesOperation.DeleteOperation] = []

    def to_json(self) -> Dict[str, Any]:
        json_dict = {"Name": self.name}
        if self._appends:
            json_dict["Appends"] = [append_op.to_json() for append_op in self._appends]
        if self._deletes:
            json_dict["Deletes"] = [delete_op.to_json() for delete_op in self._deletes]
        return json_dict

    def append(self, append_operation: AppendOperation) -> None:
        if self._appends is None:
            self._appends = set()

        if append_operation in self._appends:
            # __eq__ override lets us discard old one by passing new one due to the same timestamps
            self._appends.discard(append_operation)

        self._appends.add(append_operation)

    def delete(self, delete_operation: DeleteOperation) -> None:
        if self._deletes is None:
            self._deletes = []

        self._deletes.append(delete_operation)


class TimeSeriesRangeResult:
    def __init__(
        self,
        from_date: datetime,
        to_date: datetime,
        entries: List[TimeSeriesEntry],
        total_results: int = None,
        includes: Optional[Dict[str, Any]] = None,
    ):
        self.from_date = from_date
        self.to_date = to_date
        self.entries = entries if entries else []
        self.total_results = total_results
        self.includes = includes

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesRangeResult:
        return cls(
            Utils.string_to_datetime(json_dict["From"]),
            Utils.string_to_datetime(json_dict["To"]),
            [TimeSeriesEntry.from_json(entry_json) for entry_json in json_dict["Entries"]],
            json_dict["TotalResults"] if "TotalResults" in json_dict else None,
            json_dict.get("Includes", None),
        )


class GetTimeSeriesOperation(IOperation[TimeSeriesRangeResult]):
    def __init__(
        self,
        doc_id: str,
        time_series: str,
        from_date: datetime.datetime = None,
        to_date: datetime.datetime = None,
        start: int = 0,
        page_size: int = int_max,
        includes: Optional[Callable[[TimeSeriesIncludeBuilder], None]] = None,
    ):
        if not doc_id or doc_id.isspace():
            raise ValueError("DocId cannot be None or empty")
        if not time_series or time_series.isspace():
            raise ValueError("Timeseries cannot be None or empty")

        self._doc_id = doc_id
        self._start = start
        self._page_size = page_size
        self._name = time_series
        self._from = from_date
        self._to = to_date
        self._includes = includes

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> "RavenCommand[TimeSeriesRangeResult]":
        return self.GetTimeSeriesCommand(
            self._doc_id, self._name, self._from, self._to, self._start, self._page_size, self._includes
        )

    class GetTimeSeriesCommand(RavenCommand[TimeSeriesRangeResult]):
        def __init__(
            self,
            doc_id: str,
            name: str,
            from_date: datetime.datetime,
            to_date: datetime.datetime,
            start: int,
            page_size: int,
            includes: Callable[[TimeSeriesIncludeBuilder], None],
        ):
            super().__init__(TimeSeriesRangeResult)
            self._doc_id = doc_id
            self._name = name
            self._start = start
            self._page_size = page_size
            self._from = from_date
            self._to = to_date
            self._includes = includes

        def create_request(self, node: ServerNode) -> requests.Request:
            path_builder = [
                node.url,
                "/databases/",
                node.database,
                "/timeseries",
                "?docId=",
                self._doc_id,
            ]

            if self._start > 0:
                path_builder.append("&start=")
                path_builder.append(str(self._start))

            if self._page_size < int_max:
                path_builder.append("&pageSize=")
                path_builder.append(str(self._page_size))

            path_builder.append("&name=")
            path_builder.append(self._name)

            if self._from is not None:
                path_builder.append("&from=")
                path_builder.append(Utils.datetime_to_string(self._from))

            if self._to is not None:
                path_builder.append("&to=")
                path_builder.append(Utils.datetime_to_string(self._to))

            if self._includes is not None:
                self.add_includes_to_request(path_builder, self._includes)

            return requests.Request("GET", "".join(path_builder))

        @staticmethod
        def add_includes_to_request(
            path_builder: List[str], includes: Callable[[TimeSeriesIncludeBuilder], None]
        ) -> None:
            include_builder = TimeSeriesIncludeBuilder(DocumentConventions.default_conventions())
            includes(include_builder)

            if include_builder._include_time_series_document:
                path_builder.append("&includeDocument=true")

            if include_builder._include_time_series_tags:
                path_builder.append("&includeTags=true")

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                return

            self.result = TimeSeriesRangeResult.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return True


class TimeSeriesDetails:
    def __init__(self, key: str, values: Dict[str, List[TimeSeriesRangeResult]]):
        self.key = key
        self.values = values

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesDetails:
        return cls(
            json_dict["Id"],
            {
                key: [TimeSeriesRangeResult.from_json(value) for value in values]
                for key, values in json_dict["Values"].items()
            },
        )


class GetMultipleTimeSeriesOperation(IOperation[TimeSeriesDetails]):
    def __init__(
        self,
        doc_id: str,
        ranges: List[AbstractTimeSeriesRange],
        start: Optional[int] = 0,
        page_size: Optional[int] = int_max,
        includes: Optional[Callable[[TimeSeriesIncludeBuilder], None]] = None,
    ):
        if not doc_id or doc_id.isspace():
            raise ValueError("DocId cannot be None or empty")
        if ranges is None:
            raise ValueError("Ranges cannot be None")
        self._doc_id = doc_id
        self._ranges = ranges
        self._start = start
        self._page_size = page_size
        self._includes = includes

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> "GetMultipleTimeSeriesOperation.GetMultipleTimeSeriesCommand":
        return self.GetMultipleTimeSeriesCommand(
            self._doc_id, self._ranges, self._start, self._page_size, self._includes
        )

    class GetMultipleTimeSeriesCommand(RavenCommand[TimeSeriesDetails]):
        def __init__(
            self,
            doc_id: str,
            ranges: List[AbstractTimeSeriesRange],
            start: Optional[int] = 0,
            page_size: Optional[int] = int_max,
            includes: Optional[Callable[[TimeSeriesIncludeBuilder], None]] = None,
        ):
            super().__init__(TimeSeriesDetails)

            if doc_id is None:
                raise ValueError("DocId cannot be None")

            self._doc_id = doc_id
            self._ranges = ranges
            self._start = start
            self._page_size = page_size
            self._includes = includes

        def create_request(self, node: ServerNode) -> requests.Request:
            path_builder = [
                node.url,
                "/databases/",
                node.database,
                "/timeseries/ranges",
                "?docId=",
                self._doc_id,
            ]

            if self._start > 0:
                path_builder.append("&start=")
                path_builder.append(str(self._start))

            if self._page_size < int_max:
                path_builder.append("&pageSize=")
                path_builder.append(str(self._page_size))

            if not self._ranges:
                raise ValueError("Ranges cannot be None or empty")

            for range_ in self._ranges:
                if not range_.name or range_.name.isspace():
                    raise ValueError("Missing name argument in TimeSeriesRange. Name cannot be None or empty")

                path_builder.append("&name=")
                path_builder.append(range_.name or "")
                path_builder.append("&from=")
                path_builder.append(Utils.datetime_to_string(range_.from_date) if range_.from_date is not None else "")
                path_builder.append("&to=")
                path_builder.append(Utils.datetime_to_string(range_.to_date) if range_.to_date is not None else "")

            if self._includes is not None:
                GetTimeSeriesOperation.GetTimeSeriesCommand.add_includes_to_request(path_builder, self._includes)

            url = "".join(path_builder)

            return requests.Request("GET", url)

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                return

            self.result = TimeSeriesDetails.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return True


class TimeSeriesBatchOperation(VoidOperation):
    def __init__(self, document_id: str, operation: TimeSeriesOperation):
        if document_id is None:
            raise ValueError("Document id cannot be None")
        if operation is None:
            raise ValueError("Operation cannot be None")

        self._document_id = document_id
        self._operation = operation

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> VoidRavenCommand:
        return self.TimeSeriesBatchCommand(self._document_id, self._operation, conventions)

    class TimeSeriesBatchCommand(VoidRavenCommand):
        def __init__(self, document_id: str, operation: TimeSeriesOperation, conventions: "DocumentConventions"):
            super().__init__()
            self._document_id = document_id
            self._operation = operation
            self._conventions = conventions

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/timeseries?docId={self._document_id}"
            request = requests.Request("POST", url)
            request.data = self._operation.to_json()
            return request

        def is_read_request(self) -> bool:
            return False


class TimeSeriesItemDetail:
    def __init__(self, name: str, number_of_entities: int, start_date: datetime.datetime, end_date: datetime.datetime):
        self.name = name
        self.number_of_entries = number_of_entities
        self.start_date = start_date
        self.end_date = end_date

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesItemDetail:
        return cls(
            json_dict["Name"],
            json_dict["NumberOfEntries"],
            Utils.string_to_datetime(json_dict["StartDate"]),
            Utils.string_to_datetime(json_dict["EndDate"]),
        )


class TimeSeriesStatistics:
    def __init__(self, document_id: str, time_series: List[TimeSeriesItemDetail]):
        self.document_id = document_id
        self.time_series = time_series

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeSeriesStatistics:
        return cls(
            json_dict["DocumentId"],
            [TimeSeriesItemDetail.from_json(ts_item_detail_json) for ts_item_detail_json in json_dict["TimeSeries"]],
        )


class GetTimeSeriesStatisticsOperation(IOperation[TimeSeriesStatistics]):
    def __init__(self, document_id: str):
        self._document_id = document_id

    @property
    def document_id(self) -> str:
        return self.document_id

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: HttpCache
    ) -> RavenCommand[TimeSeriesStatistics]:
        return self.GetTimeSeriesStatisticsCommand(self._document_id)

    class GetTimeSeriesStatisticsCommand(RavenCommand[TimeSeriesStatistics]):
        def __init__(self, document_id: str):
            super().__init__()
            self._document_id = document_id

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "GET", f"{node.url}/databases/{node.database}/timeseries/stats?docId={self._document_id}"
            )

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            self.result = TimeSeriesStatistics.from_json(json.loads(response))


class ConfigureTimeSeriesOperation(MaintenanceOperation[ConfigureTimeSeriesOperationResult]):
    def __init__(self, configuration: TimeSeriesConfiguration):
        if not configuration:
            raise ValueError("Configuration cannot be None")

        self._configuration = configuration

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[ConfigureTimeSeriesOperationResult]":
        return self.ConfigureTimeSeriesCommand(self._configuration)

    class ConfigureTimeSeriesCommand(RavenCommand[ConfigureTimeSeriesOperationResult], RaftCommand):
        def __init__(self, configuration: TimeSeriesConfiguration):
            super().__init__(ConfigureTimeSeriesOperationResult)
            self._configuration = configuration

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            request = requests.Request("POST", f"{node.url}/databases/{node.database}/admin/timeseries/config")
            request.data = self._configuration.to_json()
            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if not response:
                self._throw_invalid_response()

            self.result = ConfigureTimeSeriesOperationResult.from_json(json.loads(response))

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



from __future__ import annotations

import datetime
import json
from typing import Optional, Dict, List, TYPE_CHECKING, Any

import requests

from ravendb.documents.indexes.definitions import IndexPriority, IndexLockMode, IndexType, IndexSourceType, IndexState
from ravendb.documents.operations.definitions import MaintenanceOperation
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.tools.utils import Size, Utils

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class DatabaseStatistics:
    def __init__(
        self,
        last_doc_etag: int = None,
        last_database_etag: int = None,
        count_of_indexes: int = None,
        count_of_documents: int = None,
        count_of_revision_documents: int = None,
        count_of_documents_conflicts: int = None,
        count_of_tombstones: int = None,
        count_of_conflicts: int = None,
        count_of_attachments: int = None,
        count_of_unique_attachments: int = None,
        count_of_counter_entries: int = None,
        count_of_time_series_segments: int = None,
        indexes: List[IndexInformation] = None,
        database_change_vector: str = None,
        database_id: str = None,
        is_64_bit: bool = None,
        pager: str = None,
        last_indexing_time: datetime.datetime = None,
        size_on_disk: Size = None,
        temp_buffers_size_on_disk: Size = None,
        number_of_transaction_merger_queue_operations: int = None,
    ):
        self.last_doc_etag = last_doc_etag
        self.last_database_etag = last_database_etag
        self.count_of_indexes = count_of_indexes
        self.count_of_documents = count_of_documents
        self.count_of_revision_documents = count_of_revision_documents
        self.count_of_documents_conflicts = count_of_documents_conflicts
        self.count_of_tombstones = count_of_tombstones
        self.count_of_conflicts = count_of_conflicts
        self.count_of_attachments = count_of_attachments
        self.count_of_unique_attachments = count_of_unique_attachments
        self.count_of_counter_entries = count_of_counter_entries
        self.count_of_time_series_segments = count_of_time_series_segments

        self.indexes = indexes

        self.database_change_vector = database_change_vector
        self.database_id = database_id
        self.is_64_bit = is_64_bit
        self.pager = pager
        self.last_indexing_time = last_indexing_time
        self.size_on_disk = size_on_disk
        self.temp_buffers_size_on_disk = temp_buffers_size_on_disk
        self.number_of_transaction_merger_queue_operations = number_of_transaction_merger_queue_operations

    @property
    def stale_indexes(self) -> List[IndexInformation]:
        return list(filter(lambda x: x.stale, self.indexes))

    @classmethod
    def from_json(cls, json_dict) -> DatabaseStatistics:
        return cls(
            json_dict.get("LastDocEtag", None),
            json_dict.get("LastDatabaseEtag", None),
            json_dict.get("CountOfIndexes", None),
            json_dict.get("CountOfDocuments", None),
            json_dict.get("CountOfRevisionDocuments", None),
            json_dict.get("CountOfDocumentsConflicts", None),
            json_dict.get("CountOfTombstones", None),
            json_dict.get("CountOfConflicts", None),
            json_dict.get("CountOfAttachments", None),
            json_dict.get("CountOfUniqueAttachments", None),
            json_dict.get("CountOfCounterEntries", None),
            json_dict.get("CountOfTimeSeriesSegments", None),
            [IndexInformation.from_json(x) for x in json_dict["Indexes"]] if "Indexes" in json_dict else None,
            json_dict.get("DatabaseChangeVector", None),
            json_dict.get("DatabaseId", None),
            json_dict.get("Is64Bit", None),
            json_dict.get("Pager", None),
            Utils.string_to_datetime(json_dict["LastIndexingTime"]) if json_dict.get("LastIndexingTime") else None,
            Size.from_json(json_dict.get("SizeOnDisk", None)),
            Size.from_json(json_dict.get("TempBuffersSizeOnDisk", None)),
            json_dict.get("NumberOfTransactionMergerQueueOperations", None),
        )


class GetStatisticsOperation(MaintenanceOperation[DatabaseStatistics]):
    def __init__(self, debug_tag: str = None, node_tag: str = None):
        self.debug_tag = debug_tag
        self.node_tag = node_tag

    def get_command(self, conventions: "DocumentConventions") -> _GetStatisticsCommand[DatabaseStatistics]:
        return self._GetStatisticsCommand(self.debug_tag, self.node_tag)

    class _GetStatisticsCommand(RavenCommand[DatabaseStatistics]):
        def __init__(self, debug_tag: Optional[str] = None, node_tag: Optional[str] = None):
            super().__init__(dict)
            self.debug_tag = debug_tag
            self.node_tag = node_tag

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "GET",
                f"{node.url}/databases/{node.database}"
                f"/stats{f'?{self.debug_tag}' if self.debug_tag is not None else ''}",
            )

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = DatabaseStatistics.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return True


class CollectionStatistics:
    def __init__(
        self,
        count_of_documents: Optional[int] = None,
        count_of_conflicts: Optional[int] = None,
        collections: Optional[Dict[str, int]] = None,
    ):
        self.count_of_documents = count_of_documents
        self.count_of_conflicts = count_of_conflicts
        self.collections = collections

    @classmethod
    def from_json(cls, json_dict: dict) -> CollectionStatistics:
        return cls(json_dict["CountOfDocuments"], json_dict["CountOfConflicts"], json_dict["Collections"])


class GetCollectionStatisticsOperation(MaintenanceOperation[CollectionStatistics]):
    def get_command(self, conventions) -> RavenCommand[CollectionStatistics]:
        return self.__GetCollectionStatisticsCommand()

    class __GetCollectionStatisticsCommand(RavenCommand[CollectionStatistics]):
        def __init__(self):
            super().__init__(CollectionStatistics)

        def create_request(self, server_node: ServerNode) -> requests.Request:
            return requests.Request("GET", f"{server_node.url}/databases/{server_node.database}/collections/stats")

        def is_read_request(self) -> bool:
            return True

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()
            self.result = CollectionStatistics.from_json(json.loads(response))


class IndexInformation:
    def __init__(
        self,
        stale: bool = None,
        lock_mode: IndexLockMode = None,
        priority: IndexPriority = None,
        index_type: IndexType = None,
        last_indexing_time: datetime.datetime = None,
        source_type: IndexSourceType = None,
        index_state: IndexState = None,
        name: str = None,
    ):
        self.stale = stale
        self.lock_mode = lock_mode
        self.priority = priority
        self.type = index_type
        self.last_indexing_time: datetime.datetime = last_indexing_time
        self.source_type = source_type
        self.state = index_state
        self.name = name

    @classmethod
    def from_json(cls, json_dict) -> IndexInformation:
        return cls(
            json_dict["IsStale"] if "IsStale" in json_dict else None,
            IndexLockMode(json_dict["LockMode"]),
            IndexPriority(json_dict["Priority"]) if "Priority" in json_dict else None,
            IndexType(json_dict["Type"]) if "Type" in json_dict else None,
            Utils.string_to_datetime(json_dict["LastIndexingTime"]) if "LastIndexingTime" in json_dict else None,
            IndexSourceType(json_dict["SourceType"]) if "SourceType" in json_dict else None,
            IndexState(json_dict["State"]) if "State" in json_dict else None,
            json_dict["Name"] if "Name" in json_dict else None,
        )


class DetailedDatabaseStatistics(DatabaseStatistics):
    def __init__(
        self,
        last_doc_etag: int = None,
        last_database_etag: int = None,
        count_of_indexes: int = None,
        count_of_documents: int = None,
        count_of_revision_documents: int = None,
        count_of_documents_conflicts: int = None,
        count_of_tombstones: int = None,
        count_of_conflicts: int = None,
        count_of_attachments: int = None,
        count_of_unique_attachments: int = None,
        count_of_counter_entries: int = None,
        count_of_time_series_segments: int = None,
        indexes: List[IndexInformation] = None,
        database_change_vector: str = None,
        database_id: str = None,
        is_64_bit: bool = None,
        pager: str = None,
        last_indexing_time: datetime.datetime = None,
        size_on_disk: Size = None,
        temp_buffers_size_on_disk: Size = None,
        number_of_transaction_merger_queue_operations: int = None,
        count_of_identities: int = None,
        count_of_compare_exchange: int = None,
        count_of_compare_exchange_tombstones: int = None,
    ):
        super().__init__(
            last_doc_etag,
            last_database_etag,
            count_of_indexes,
            count_of_documents,
            count_of_revision_documents,
            count_of_documents_conflicts,
            count_of_tombstones,
            count_of_conflicts,
            count_of_attachments,
            count_of_unique_attachments,
            count_of_counter_entries,
            count_of_time_series_segments,
            indexes,
            database_change_vector,
            database_id,
            is_64_bit,
            pager,
            last_indexing_time,
            size_on_disk,
            temp_buffers_size_on_disk,
            number_of_transaction_merger_queue_operations,
        )

        self.count_of_identities = count_of_identities
        self.count_of_compare_exchange = count_of_compare_exchange
        self.count_of_compare_exchange_tombstones = count_of_compare_exchange_tombstones

    @classmethod
    def from_json(cls, json_dict: Dict) -> DetailedDatabaseStatistics:
        database_stats = DatabaseStatistics.from_json(json_dict)

        detailed_database_stats = cls()
        detailed_database_stats.__dict__.update(database_stats.__dict__)

        detailed_database_stats.count_of_identities = json_dict["CountOfIdentities"]
        detailed_database_stats.count_of_compare_exchange = json_dict["CountOfCompareExchange"]
        detailed_database_stats.count_of_compare_exchange_tombstones = json_dict["CountOfCompareExchangeTombstones"]

        return detailed_database_stats


class GetDetailedStatisticsOperation(MaintenanceOperation[DetailedDatabaseStatistics]):
    def __init__(self, debug_tag: str = None):
        self.__debug_tag = debug_tag

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[DetailedDatabaseStatistics]:
        return self.DetailedDatabaseStatisticsCommand(self.__debug_tag)

    class DetailedDatabaseStatisticsCommand(RavenCommand[DetailedDatabaseStatistics]):
        def __init__(self, debug_tag: str):
            super().__init__(DetailedDatabaseStatistics)
            self.__debug_tag = debug_tag

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/stats/detailed"
            if self.__debug_tag is not None:
                url += f"?{self.__debug_tag}"

            return requests.Request("GET", url)

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = DetailedDatabaseStatistics.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return True


class CollectionDetails:
    def __init__(
        self,
        name: str = None,
        count_of_documents: int = None,
        size: Size = None,
        documents_size: Size = None,
        tombstones_size: Size = None,
        revisions_size: Size = None,
    ):
        self.name = name
        self.count_of_documents = count_of_documents
        self.size = size
        self.documents_size = documents_size
        self.tombstones_size = tombstones_size
        self.revisions_size = revisions_size

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> CollectionDetails:
        return cls(
            json_dict["Name"],
            json_dict["CountOfDocuments"],
            Size.from_json(json_dict["Size"]),
            Size.from_json(json_dict["DocumentsSize"]),
            Size.from_json(json_dict["TombstonesSize"]),
            Size.from_json(json_dict["RevisionsSize"]),
        )


class DetailedCollectionStatistics:
    def __init__(
        self,
        count_of_documents: int = None,
        count_of_conflicts: int = None,
        collections: Dict[str, CollectionDetails] = None,
    ) -> None:
        self.count_of_documents = count_of_documents
        self.count_of_conflicts = count_of_conflicts
        self.collections = collections

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "DetailedCollectionStatistics":
        return cls(
            json_dict["CountOfDocuments"],
            json_dict["CountOfConflicts"],
            {key: CollectionDetails.from_json(value) for key, value in json_dict["Collections"].items()},
        )


class GetDetailedCollectionStatisticsOperation(MaintenanceOperation[DetailedCollectionStatistics]):
    def get_command(
        self, conventions: "DocumentConventions"
    ) -> "GetDetailedCollectionStatisticsOperation.GetDetailedCollectionStatisticsCommand":
        return self.GetDetailedCollectionStatisticsCommand()

    class GetDetailedCollectionStatisticsCommand(RavenCommand[DetailedCollectionStatistics]):
        def __init__(self):
            super().__init__(DetailedCollectionStatistics)

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("GET", f"{node.url}/databases/{node.database}/collections/stats/detailed")

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if not response:
                self._throw_invalid_response()

            self.result = DetailedCollectionStatistics.from_json(json.loads(response))



from typing import Optional

from ravendb.documents.operations.connection_strings import ConnectionString
import ravendb.serverwide
from ravendb.documents.operations.etl.configuration import EtlConfiguration


class SqlConnectionString(ConnectionString):
    def __init__(self, name: str, connection_string: Optional[str] = None, factory_name: Optional[str] = None):
        super().__init__(name)
        self.connection_string = connection_string
        self.factory_name = factory_name

    @property
    def get_type(self):
        return ravendb.serverwide.ConnectionStringType.SQL


# todo: implement
class SqlEtlConfiguration(EtlConfiguration[SqlConnectionString]):
    pass



from typing import List, Dict, Set

from ravendb.documents.queries.query import QueryResult


class Highlightings:
    def __init__(self, field_name: str):
        self.__field_name = field_name
        self.__highlightings: Dict[str, List[str]] = {}

    @property
    def field_name(self) -> str:
        return self.__field_name

    @property
    def result_indents(self) -> Set[str]:
        return set(self.__highlightings.keys())

    def get_fragments(self, key: str) -> List[str]:
        return self.__highlightings.get(key, None) or []

    def update(self, highlightings: Dict[str, Dict[str, List[str]]]) -> None:
        self.__highlightings.clear()

        if highlightings is None or self.field_name not in highlightings:
            return

        result = highlightings.get(self.field_name)
        for key, value in result.items():
            self.__highlightings[key] = value


class QueryHighlightings:
    def __init__(self):
        self.__highlightings: List[Highlightings] = []

    def add(self, field_name: str) -> Highlightings:
        field_highlightings = Highlightings(field_name)
        self.__highlightings.append(field_highlightings)
        return field_highlightings

    def update(self, query_result: QueryResult) -> None:
        for field_highlightings in self.__highlightings:
            field_highlightings.update(query_result.highlightings)


class HighlightingOptions:
    def __init__(self, group_key: str = None, pre_tags: List[str] = None, post_tags: List[str] = None):
        self.group_key = group_key
        self.pre_tags = pre_tags
        self.post_tags = post_tags

    def to_json(self) -> dict:
        return {"GroupKey": self.group_key, "PreTags": self.pre_tags, "PostTags": self.post_tags}



from typing import Union, Dict

import requests

from ravendb.primitives import constants


class HttpExtensions:
    @staticmethod
    def get_required_etag_header(response: requests.Response) -> str:
        if constants.Headers.ETAG in response.headers:
            headers = response.headers[constants.Headers.ETAG]
            if headers:
                return HttpExtensions.etag_header_to_change_vector(headers)

    @staticmethod
    def get_etag_header(response_or_headers: Union[requests.Response, Dict[str, str]]) -> Union[None, str]:
        headers = (
            response_or_headers.headers if isinstance(response_or_headers, requests.Response) else response_or_headers
        )
        if constants.Headers.ETAG in headers:
            headers = headers.get(constants.Headers.ETAG)
            if headers:
                return HttpExtensions.etag_header_to_change_vector(headers)
        return None

    @staticmethod
    def etag_header_to_change_vector(response_header: str) -> str:
        if not response_header:
            raise ValueError("Response didn't had an ETag header")
        if response_header.startswith('"'):
            return response_header[1:-1]
        return response_header

    @staticmethod
    def get_boolean_header(response: requests.Response, header: str) -> Union[None, bool]:
        if header in response.headers:
            first_header = response.headers.get(header)
            return bool(first_header)
        return None



from __future__ import annotations

import json
import logging
from datetime import datetime
from typing import Union, List, Type, TypeVar, Optional, Dict, Any, TYPE_CHECKING

import requests

from ravendb.http.server_node import ServerNode
from ravendb.documents.commands.multi_get import GetRequest, MultiGetCommand
from ravendb.documents.commands.revisions import GetRevisionsCommand
from ravendb.documents.session.document_info import DocumentInfo

from ravendb.primitives import constants
from ravendb.tools.utils import Utils, CaseInsensitiveDict
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
from ravendb.http.raven_command import RavenCommand
from ravendb.documents.commands.crud import GetDocumentsCommand

if TYPE_CHECKING:
    from ravendb.json.result import JsonArrayResult
    from ravendb.documents.session.document_session_operations.in_memory_document_session_operations import (
        InMemoryDocumentSessionOperations,
    )
    from ravendb.documents.commands.crud import GetDocumentsResult

_T = TypeVar("_T")


class MultiGetOperation:
    def __init__(self, session: InMemoryDocumentSessionOperations):
        self.__session = session

    def create_request(self, requests: List[GetRequest]):
        return MultiGetCommand(self.__session.advanced.request_executor, requests)

    def set_result(self, result: dict) -> None:
        pass


class LoadStartingWithOperation:
    logger = logging.getLogger("load_starting_with_operation")

    def __init__(self, session: InMemoryDocumentSessionOperations):
        self.__session = session

        self.__start_with: Union[None, str] = None
        self.__matches: Union[None, str] = None
        self.__start: Union[None, int] = None
        self.__page_size: Union[None, int] = None
        self.__exclude: Union[None, str] = None
        self.__start_after: Union[None, str] = None

        self.__returned_ids: List[str] = []
        self.__results_set: bool = False
        self.__results: Union[None, GetDocumentsResult] = None

    def with_start_with(
        self, id_prefix: str, matches: str = None, start: int = 0, page_size: int = 25, exclude=None, start_after=None
    ):
        self.__start_with = id_prefix
        self.__matches = matches
        self.__start = start
        self.__page_size = page_size
        self.__exclude = exclude
        self.__start_after = start_after

    def create_request(self):
        self.__session.increment_requests_count()

        self.logger.info(
            f"Requesting documents with ids starting with '{self.__start_with}' from {self.__session.advanced.store_identifier}"
        )

        return GetDocumentsCommand.from_starts_with(
            self.__start_with, self.__start_after, self.__matches, self.__exclude, self.__start, self.__page_size, False
        )

    def set_result(self, result: GetDocumentsResult) -> None:
        self.__results_set = True

        if self.__session.no_tracking:
            self.__results = result
            return

        for document in result.results:
            if not document:
                continue
            new_document_info = DocumentInfo.get_new_document_info(document)
            self.__session._documents_by_id.add(new_document_info)
            self.__returned_ids.append(new_document_info.key)

    def get_documents(self, object_type: Type[_T]) -> List[_T]:
        i = 0
        final_results = []

        if self.__session.no_tracking:
            if self.__results is None:
                raise RuntimeError("Cannot execute get_documents before operation execution")

            if not self.__results:
                return []

            final_results = [
                self.__session.track_entity_document_info(object_type, DocumentInfo.get_new_document_info(document))
                for document in self.__results.results
            ]
        else:
            final_results = [self.__get_document(object_type, key) for key in self.__returned_ids]

        return final_results

    def __get_document(self, object_type: Type[_T], key: str) -> _T:
        if not key:
            return None

        if self.__session.is_deleted(key):
            return None

        doc = self.__session._documents_by_id.get(key)
        if doc:
            return self.__session.track_entity_document_info(object_type, doc)

        return None


class GetRevisionOperation:
    def __init__(self, session: InMemoryDocumentSessionOperations = None):
        if session is None:
            raise ValueError("Session cannot be None")
        self._session = session
        self._command: Optional[GetRevisionsCommand] = None
        self._result: Optional[JsonArrayResult] = None

    @classmethod
    def from_start_page(
        cls,
        session: InMemoryDocumentSessionOperations,
        id_: str,
        start: int,
        page_size: int,
        metadata_only: bool = False,
    ) -> GetRevisionOperation:
        self = cls(session)
        if id_ is None:
            raise ValueError("Id cannot be None")

        self._session = session
        self._command = GetRevisionsCommand.from_start_page(id_, start, page_size, metadata_only)
        return self

    @classmethod
    def from_before_date(
        cls, session: InMemoryDocumentSessionOperations, id_: str, before: datetime
    ) -> GetRevisionOperation:
        self = cls(session)
        self._command = GetRevisionsCommand.from_before(id_, before)
        return self

    @classmethod
    def from_change_vector(cls, session: InMemoryDocumentSessionOperations, change_vector: str) -> GetRevisionOperation:
        self = cls(session)
        self._command = GetRevisionsCommand.from_change_vector(change_vector)
        return self

    @classmethod
    def from_change_vectors(
        cls, session: InMemoryDocumentSessionOperations, change_vectors: List[str]
    ) -> GetRevisionOperation:
        self = cls(session)
        self._command = GetRevisionsCommand.from_change_vectors(change_vectors)
        return self

    def create_request(self) -> GetRevisionsCommand:
        if self._command.change_vectors is not None:
            return (
                None
                if self._session.check_if_all_change_vectors_are_already_included(self._command.change_vectors)
                else self._command
            )

        if self._command.change_vector is not None:
            return (
                None
                if self._session.check_if_all_change_vectors_are_already_included([self._command.change_vector])
                else self._command
            )

        if self._command.before is not None:
            return (
                None
                if self._session.check_if_revisions_by_date_time_before_already_included(
                    self._command.id, self._command.before
                )
                else self._command
            )

        return self._command

    def set_result(self, result: JsonArrayResult) -> None:
        self._result = result

    @property
    def command(self) -> GetRevisionsCommand:
        return self._command

    def _get_revision(self, document: Dict[str, Any], object_type: Type[_T] = dict) -> _T:
        if document is None:
            return Utils.get_default_value(object_type)

        metadata = None
        id_ = None
        if constants.Documents.Metadata.KEY in document:
            metadata = document.get(constants.Documents.Metadata.KEY)
            id_node = metadata.get(constants.Documents.Metadata.ID, None)
            if id_node is not None:
                id_ = id_node

        change_vector = None
        if metadata is not None and constants.Documents.Metadata.CHANGE_VECTOR in metadata:
            change_vector_node = metadata.get(constants.Documents.Metadata.CHANGE_VECTOR, None)
            if change_vector_node is not None:
                change_vector = change_vector_node

        entity = self._session.entity_to_json.convert_to_entity(
            object_type, id_, document, not self._session.no_tracking
        )
        document_info = DocumentInfo()
        document_info.key = id_
        document_info.change_vector = change_vector
        document_info.document = document
        document_info.metadata = metadata
        document_info.entity = entity
        self._session.documents_by_entity[entity] = document_info

        return entity

    def get_revisions_for(self, object_type: Type[_T]) -> List[_T]:
        results_count = len(self._result.results)
        results = []
        for i in range(results_count):
            document = self._result.results[i]
            results.append(self._get_revision(document, object_type))

        return results

    def get_revisions_metadata_for(self) -> List[MetadataAsDictionary]:
        results_count = len(self._result.results)
        results = []
        for i in range(results_count):
            document = self._result.results[i]

            metadata = None
            if constants.Documents.Metadata.KEY in document:
                metadata = document.get(constants.Documents.Metadata.KEY)

            results.append(MetadataAsDictionary(metadata))

        return results

    def get_revision(self, object_type: Type[_T]) -> _T:
        if self._result is None:
            revision: DocumentInfo

            if self._command.change_vectors is not None:
                for change_vector in self._command.change_vectors:
                    revision = self._session.include_revisions_by_change_vector.get(change_vector, None)
                    if revision is not None:
                        return self._get_revision(revision.document, object_type)

            if self._command.change_vector is not None and self._session.include_revisions_by_change_vector is not None:
                revision = self._session.include_revisions_by_change_vector.get(self._command.change_vector, None)
                if revision is not None:
                    return self._get_revision(revision.document, object_type)

            if self._command.before is not None and self._session.include_revisions_by_date_time_before is not None:
                dictionary_date_time_to_document = self._session.include_revisions_by_date_time_before.get(
                    self.command.id
                )
                if dictionary_date_time_to_document is not None:
                    revision = dictionary_date_time_to_document.get(self._command.before)
                    if revision is not None:
                        return self._get_revision(revision.document, object_type)

            return Utils.get_default_value(object_type)

        document = self._result.results[0]
        return self._get_revision(document, object_type)

    def get_revisions(self, object_type: Type[_T]) -> Dict[str, _T]:
        results = CaseInsensitiveDict()

        if self._result is None:
            for change_vector in self._command.change_vectors:
                revision = self._session.include_revisions_by_change_vector.get(change_vector)
                if revision is not None:
                    results[change_vector] = self._get_revision(revision.document, object_type)

            return results

        for i in range(len(self._command.change_vectors)):
            change_vector = self._command.change_vectors[i]
            if change_vector is None:
                continue

            json_dict = self._result.results[i]
            results[change_vector] = self._get_revision(json_dict, object_type)

        return results


class DocumentRevisionsCount:
    def __init__(self, revisions_count: int = None):
        self.revisions_count = revisions_count

    @classmethod
    def from_json(cls, json_dict: Dict) -> DocumentRevisionsCount:
        return cls(json_dict["RevisionsCount"])


class GetRevisionsCountOperation:
    def __init__(self, doc_id: str):
        self._doc_id = doc_id

    def create_request(self) -> RavenCommand[int]:
        return self.GetRevisionsCountCommand(self._doc_id)

    class GetRevisionsCountCommand(RavenCommand[int]):
        def __init__(self, id_: str):
            super().__init__(int)
            if id_ is None:
                raise ValueError("Id cannot be None")

            self._id = id_

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "GET", f"{node.url}/databases/{node.database}/revisions/count?&id={Utils.quote_key(self._id)}"
            )

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self.result = 0
                return

            self.result = DocumentRevisionsCount.from_json(json.loads(response)).revisions_count

        def is_read_request(self) -> bool:
            return True



from __future__ import annotations

import enum
from typing import TYPE_CHECKING, TypeVar, Optional

from ravendb.http.request_executor import ClusterRequestExecutor
from ravendb.http.topology import Topology
from ravendb.serverwide.operations.common import (
    GetBuildNumberOperation,
    ServerOperation,
    ServerWideOperation,
)
from ravendb.tools.utils import CaseInsensitiveDict

if TYPE_CHECKING:
    from ravendb.documents.store.definition import DocumentStore
    from ravendb.documents.operations.operation import Operation
    from ravendb.documents.operations.definitions import OperationIdResult

_T_OperationResult = TypeVar("_T_OperationResult")


class ConnectionStringType(enum.Enum):
    NONE = "NONE"
    RAVEN = "RAVEN"
    SQL = "SQL"
    OLAP = "OLAP"


class ServerOperationExecutor:
    def __init__(
        self,
        store: DocumentStore,
        request_executor: ClusterRequestExecutor,
        initial_request_executor: ClusterRequestExecutor = None,
        cache: CaseInsensitiveDict = None,
        node_tag: str = None,
    ):
        if store is None:
            raise ValueError("Store cannot be None")
        if request_executor is None:
            raise ValueError("Request Executor cannot be None")
        if cache is None:
            cache = CaseInsensitiveDict()

        self._store = store
        self._request_executor = request_executor
        self._initial_request_executor = initial_request_executor
        self._node_tag = node_tag
        self._cache = cache

        store.register_events_for_request_executor(self._request_executor)

        if self._node_tag is None:
            self._store.add_after_close(lambda: self._request_executor.close())

    @classmethod
    def from_store(cls, store: DocumentStore):
        return cls(store, cls.create_request_executor(store), None, CaseInsensitiveDict(), None)

    def for_node(self, node_tag: str) -> ServerOperationExecutor:
        if not node_tag or node_tag.isspace():
            raise ValueError("Value cannot be None or whitespace")

        if self._node_tag and self._node_tag.lower() == node_tag.lower():
            return self

        if self._store.conventions.disable_topology_updates:
            raise RuntimeError(
                "Cannot switch server operation executor, because conventions.disable_topology_updates is set to 'True'"
            )

        if node_tag in self._cache:
            return self._cache[node_tag]

        request_executor = self._initial_request_executor or self._request_executor
        topology = self._get_topology(self._request_executor)

        node = next((node for node in topology.nodes if node_tag.lower() == node.cluster_tag.lower()), None)

        if node is None:
            available_nodes = str.join(", ", [node.cluster_tag for node in topology.nodes])
            raise RuntimeError(f"Could not find node '{node_tag}' in the topology. Available nodes: {available_nodes}")

        cluster_executor = ClusterRequestExecutor.create_for_single_node(
            node.url,
            self._store.thread_pool_executor,
            self._store.conventions,
            self._store.certificate_pem_path,
            self._store.trust_store_path,
        )
        return ServerOperationExecutor(self._store, cluster_executor, request_executor, self._cache, node.cluster_tag)

    def send(self, operation: ServerOperation[_T_OperationResult]) -> Optional[_T_OperationResult]:
        command = operation.get_command(self._request_executor.conventions)
        self._request_executor.execute_command(command)

        if isinstance(operation, ServerOperation):
            return command.result

    def send_async(self, operation: ServerOperation[OperationIdResult]) -> Operation:
        command = operation.get_command(self._request_executor.conventions)

        self._request_executor.execute_command(command)
        return ServerWideOperation(
            self._request_executor,
            self._request_executor.conventions,
            command.result.operation_id,
            command.selected_node_tag if command.selected_node_tag else command.result.operation_node_tag,
        )

    def close(self) -> None:
        if self._node_tag is not None:
            return

        if self._request_executor is not None:
            self._request_executor.close()

        cache = self._cache
        if cache is not None:
            for key, value in cache.items():
                request_executor = value._request_executor
                if request_executor is not None:
                    request_executor.close()

            cache.clear()

    def _get_topology(self, request_executor: ClusterRequestExecutor) -> Topology:
        topology: Optional[Topology] = None
        try:
            topology = request_executor.topology
            if topology is None:
                # a bit rude way to make sure that topology was refreshed
                # but it handles a case when first topology update failed

                operation = GetBuildNumberOperation()
                command = operation.get_command(request_executor.conventions)
                request_executor.execute_command(command)

                topology = request_executor.topology

        except Exception:
            ...  # ignored

        if topology is None:
            raise RuntimeError("Could not fetch the topology")

        return topology

    @staticmethod
    def create_request_executor(store: DocumentStore) -> ClusterRequestExecutor:
        return (
            ClusterRequestExecutor.create_for_single_node(
                store.urls[0],
                store.thread_pool_executor,
                store.conventions,
                store.certificate_pem_path,
                store.trust_store_path,
            )
            if store.conventions.disable_topology_updates
            else ClusterRequestExecutor.create_without_database_name(
                store.urls,
                store.thread_pool_executor,
                store.conventions,
                store.certificate_pem_path,
                store.trust_store_path,
            )
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return



from __future__ import annotations
from datetime import datetime
from typing import List, Dict


class User:
    def __init__(
        self,
        Id: str = None,
        name: str = None,
        last_name: str = None,
        address_id: str = None,
        count: int = None,
        age: int = None,
    ):
        self.Id = Id
        self.name = name
        self.last_name = last_name
        self.address_id = address_id
        self.count = count
        self.age = age


class Post:
    def __init__(
        self,
        Id: str = None,
        title: str = None,
        desc: str = None,
        comments: List[Post] = None,
        attachment_ids: str = None,
        created_at: datetime.date = None,
    ):
        self.Id = Id
        self.title = title
        self.desc = desc
        self.comments = comments
        self.attachment_ids = attachment_ids
        self.created_at = created_at



from typing import Dict, List

import ravendb.primitives.constants as constants
from ravendb.documents.session.document_info import DocumentInfo
from ravendb.documents.session.misc import DocumentsChanges


class JsonOperation:
    @staticmethod
    def entity_changed(new_obj: dict, document_info: DocumentInfo, changes: Dict[str, List[DocumentsChanges]]) -> bool:
        doc_changes = [] if changes is not None else None

        if not document_info.new_document and document_info.entity:
            return JsonOperation.compare_json(
                "", document_info.key, document_info.document, new_obj, changes, doc_changes
            )

        if changes is None:
            return True

        JsonOperation.new_change(None, None, None, None, doc_changes, DocumentsChanges.ChangeType.DOCUMENT_ADDED)
        changes.update({document_info.key: doc_changes})
        return True

    @staticmethod
    def new_change(field_path, name, new_value, old_value, doc_changes, change_type):
        changes = {
            "old_value": old_value,
            "new_value": new_value,
            "change": change_type,
            "field_name": name,
            "field_path": field_path,
        }
        doc_changes.append(changes)

    @staticmethod
    def compare_json(
        field_path: str,
        key: str,
        original_json: Dict,
        new_json: Dict,
        changes: Dict[str, List[DocumentsChanges]],
        doc_changes: List[DocumentsChanges],
    ) -> bool:
        old_json_props = set(original_json.keys())
        new_json_props = set(new_json.keys())

        # todo: Discuss about that condition - add/del Id causes changes
        if "Id" in old_json_props:
            old_json_props.remove("Id")
        if "Id" in new_json_props:
            new_json_props.remove("Id")

        new_fields = new_json_props - old_json_props
        removed_fields = old_json_props - new_json_props

        for field in removed_fields:
            if changes is None:
                return True
            JsonOperation.new_change(
                field_path, field, None, None, doc_changes, DocumentsChanges.ChangeType.REMOVED_FIELD
            )

        for prop in new_json_props:
            if (
                prop == constants.Documents.Metadata.LAST_MODIFIED
                or prop == constants.Documents.Metadata.COLLECTION
                or prop == constants.Documents.Metadata.CHANGE_VECTOR
                or prop == constants.Documents.Metadata.ID
            ):
                continue

            if prop in new_fields:
                if changes is None:
                    return True
                JsonOperation.new_change(
                    field_path, prop, new_json[prop], None, doc_changes, DocumentsChanges.ChangeType.NEW_FIELD
                )
                continue

            new_prop = new_json[prop]
            old_prop = original_json[prop]

            if isinstance(new_prop, (int, float, bool, str)):
                if new_prop == old_prop or JsonOperation.compare_values(old_prop, new_prop):
                    continue
                if changes is None:
                    return True
                JsonOperation.new_change(
                    field_path, prop, new_prop, old_prop, doc_changes, DocumentsChanges.ChangeType.FIELD_CHANGED
                )

            elif new_prop is None:
                if old_prop is None:
                    continue
                if changes is None:
                    return True
                JsonOperation.new_change(
                    field_path, prop, None, old_prop, doc_changes, DocumentsChanges.ChangeType.FIELD_CHANGED
                )

            elif isinstance(new_prop, (list, set)):
                if not isinstance(old_prop, (list, set)):
                    if changes is None:
                        return True
                    JsonOperation.new_change(
                        field_path, prop, new_prop, old_prop, doc_changes, DocumentsChanges.ChangeType.FIELD_CHANGED
                    )
                    continue
                changed = JsonOperation.compare_json_array(
                    JsonOperation.field_path_combine(field_path, prop),
                    key,
                    old_prop,
                    new_prop,
                    changes,
                    doc_changes,
                    prop,
                )
                if changes is None and changed is True:
                    return True
            else:
                if old_prop is None:
                    if changes is None:
                        return True
                    JsonOperation.new_change(
                        field_path, prop, new_prop, None, doc_changes, DocumentsChanges.ChangeType.FIELD_CHANGED
                    )
                changed = JsonOperation.compare_json(
                    JsonOperation.field_path_combine(field_path, prop), key, old_prop, new_prop, changes, doc_changes
                )
                if changes is None and changed is True:
                    return True
        if changes is None or len(doc_changes) <= 0:
            return False
        changes.update({key: doc_changes})
        return True

    @staticmethod
    def compare_values(old_prop, new_prop):
        return old_prop == new_prop and type(old_prop) == type(new_prop)

    @staticmethod
    def field_path_combine(path1, path2):
        return path2 if not path1 else f"{path1}.{path2}"

    @staticmethod
    def add_index_field_path(field_path, position):
        return f"{field_path}[{position}]"

    @staticmethod
    def compare_json_array(field_path: str, key: str, old_collection, new_collection, changes, doc_changes, prop_name):
        # if we don't care about the changes
        if len(old_collection) != len(new_collection) and changes is None:
            return True
        position = 0
        changed = False

        while position < len(old_collection) and position < len(new_collection):
            old_collection_item = old_collection[position]
            new_collection_item = new_collection[position]
            if old_collection_item is None:
                if new_collection_item is not None:
                    changed = True
                    if changes is not None:
                        JsonOperation.new_change(
                            JsonOperation.add_index_field_path(field_path, position),
                            prop_name,
                            new_collection_item,
                            old_collection_item,
                            doc_changes,
                            DocumentsChanges.ChangeType.ARRAY_VALUE_ADDED,
                        )
            elif isinstance(old_collection_item, (list, set)):
                if isinstance(new_collection_item, (list, set)):
                    changed |= JsonOperation.compare_json_array(
                        JsonOperation.add_index_field_path(field_path, position),
                        key,
                        old_collection_item,
                        new_collection_item,
                        changes,
                        doc_changes,
                        prop_name,
                    )
                else:
                    changed = True
                    if changes is None:
                        JsonOperation.new_change(
                            JsonOperation.add_index_field_path(field_path, position),
                            prop_name,
                            new_collection_item,
                            old_collection_item,
                            doc_changes,
                            DocumentsChanges.ChangeType.ARRAY_VALUE_CHANGED,
                        )
            elif isinstance(new_collection_item, (int, float, bool, str)):
                if not str(old_collection_item) == str(new_collection_item):
                    if changes is not None:
                        JsonOperation.new_change(
                            JsonOperation.add_index_field_path(field_path, position),
                            prop_name,
                            new_collection_item,
                            old_collection_item,
                            doc_changes,
                            DocumentsChanges.ChangeType.ARRAY_VALUE_CHANGED,
                        )
                    changed = True
            else:
                if type(old_collection_item) is type(new_collection_item):
                    changed |= JsonOperation.compare_json(
                        JsonOperation.add_index_field_path(field_path, position),
                        key,
                        old_collection_item,
                        new_collection_item,
                        changes,
                        doc_changes,
                    )
                else:
                    changed = True
                    if changes is not None:
                        JsonOperation.new_change(
                            JsonOperation.add_index_field_path(field_path, position),
                            prop_name,
                            new_collection_item,
                            old_collection_item,
                            doc_changes,
                            DocumentsChanges.ChangeType.ARRAY_VALUE_CHANGED,
                        )
            position += 1
        if changes is None:
            return changed

        # if one of the arrays is larger than the other
        while position < len(old_collection):
            old_collection_item = old_collection[position]
            JsonOperation.new_change(
                field_path,
                prop_name,
                None,
                old_collection_item,
                doc_changes,
                DocumentsChanges.ChangeType.ARRAY_VALUE_REMOVED,
            )
            position += 1

        while position < len(new_collection):
            new_collection_item = new_collection[position]
            JsonOperation.new_change(
                field_path,
                prop_name,
                new_collection_item,
                None,
                doc_changes,
                DocumentsChanges.ChangeType.ARRAY_VALUE_ADDED,
            )
            position += 1

        return changed



from __future__ import annotations
import datetime
import json
from enum import Enum
from typing import Union, Optional, Dict, Generic, TYPE_CHECKING

import requests

from ravendb.documents.conventions import DocumentConventions
from ravendb.documents.operations.definitions import OperationIdResult, IOperation, _Operation_T
from ravendb.documents.operations.misc import QueryOperationOptions
from ravendb.documents.queries.index_query import IndexQuery
from ravendb.http.http_cache import HttpCache
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.tools.utils import Utils

if TYPE_CHECKING:
    from ravendb.documents.store.definition import DocumentStore


class PatchRequest:
    def __init__(self, script: Optional[str] = "", values: Optional[Dict[str, object]] = None):
        self.values: Dict[str, object] = values or {}
        self.script = script

    @classmethod
    def for_script(cls, script: str) -> PatchRequest:
        request = cls()
        request.script = script
        return request

    def serialize(self):
        return {"Script": self.script, "Values": self.values}


class PatchStatus(Enum):
    DOCUMENT_DOES_NOT_EXIST = "DocumentDoesNotExist"
    CREATED = "Created"
    PATCHED = "Patched"
    SKIPPED = "Skipped"
    NOT_MODIFIED = "NotModified"

    def __str__(self):
        return self.value


class PatchResult:
    def __init__(
        self,
        status: Optional[PatchStatus] = None,
        modified_document: Optional[dict] = None,
        original_document: Optional[dict] = None,
        debug: Optional[dict] = None,
        last_modified: Optional[datetime.datetime] = None,
        change_vector: Optional[str] = None,
        collection: Optional[str] = None,
    ):
        self.status = status
        self.modified_document = modified_document
        self.original_document = original_document
        self.debug = debug
        self.last_modified = last_modified
        self.change_vector = change_vector
        self.collection = collection

    @classmethod
    def from_json(cls, json_dict: dict) -> PatchResult:
        return cls(
            PatchStatus(json_dict["Status"]),
            json_dict.get("ModifiedDocument", None),
            json_dict.get("OriginalDocument", None),
            json_dict.get("Debug", None),
            Utils.string_to_datetime(json_dict["LastModified"]) if "LastModified" in json_dict else None,
            json_dict.get("ChangeVector", None),
            json_dict.get("Collection", None),
        )


class PatchOperation(IOperation[PatchResult]):
    class Payload:
        def __init__(self, patch: PatchRequest, patch_if_missing: PatchResult):
            self._patch = patch
            self._patch_if_missing = patch_if_missing

        @property
        def patch(self):
            return self._patch

        @property
        def patch_if_missing(self):
            return self._patch_if_missing

    class Result(Generic[_Operation_T]):
        def __init__(self, status: PatchStatus = None, document: _Operation_T = None):
            self.status = status
            self.document = document

    def __init__(
        self,
        key: str,
        change_vector: str,
        patch: PatchRequest,
        patch_if_missing: Optional[PatchRequest] = None,
        skip_patch_if_change_vector_mismatch: Optional[bool] = None,
    ):
        if not patch:
            raise ValueError("Patch cannot be None")

        if patch.script.isspace():
            raise ValueError("Patch script cannot be None or whitespace")

        if patch_if_missing and patch_if_missing.script.isspace():
            raise ValueError("PatchIfMissing script cannot be None or whitespace")

        self._key = key
        self._change_vector = change_vector
        self._patch = patch
        self._patch_if_missing = patch_if_missing
        self._skip_patch_if_change_vector_mismatch = skip_patch_if_change_vector_mismatch

    def get_command(
        self,
        store: "DocumentStore",
        conventions: DocumentConventions,
        cache: HttpCache,
        return_debug_information: Optional[bool] = False,
        test: Optional[bool] = False,
    ) -> RavenCommand[_Operation_T]:
        return self.PatchCommand(
            conventions,
            self._key,
            self._change_vector,
            self._patch,
            self._patch_if_missing,
            self._skip_patch_if_change_vector_mismatch,
            return_debug_information,
            test,
        )

    class PatchCommand(RavenCommand[PatchResult]):
        def __init__(
            self,
            conventions: "DocumentConventions",
            document_id: str,
            change_vector: str,
            patch: PatchRequest,
            patch_if_missing: PatchRequest,
            skip_patch_if_change_vector_mismatch: Optional[bool] = False,
            return_debug_information: Optional[bool] = False,
            test: Optional[bool] = False,
        ):
            if conventions is None:
                raise ValueError("None conventions are invalid")
            if document_id is None:
                raise ValueError("None key is invalid")
            if patch is None:
                raise ValueError("None patch is invalid")
            if patch_if_missing and not patch_if_missing.script:
                raise ValueError("None or Empty script is invalid")
            if patch and patch.script.isspace():
                raise ValueError("None or empty patch_if_missing.script is invalid")
            if not document_id:
                raise ValueError("Document_id canoot be None")

            super().__init__(PatchResult)
            self._conventions = conventions
            self._document_id = document_id
            self._change_vector = change_vector
            self._patch = patch
            self._patch_if_missing = patch_if_missing
            self._skip_patch_if_change_vector_mismatch = skip_patch_if_change_vector_mismatch
            self._return_debug_information = return_debug_information
            self._test = test

        def create_request(self, server_node: ServerNode) -> requests.Request:
            path = f"docs?id={Utils.quote_key(self._document_id)}"
            if self._skip_patch_if_change_vector_mismatch:
                path += "&skipPatchIfChangeVectorMismatch=true"
            if self._return_debug_information:
                path += "&debug=true"
            if self._test:
                path += "&test=true"
            url = f"{server_node.url}/databases/{server_node.database}/{path}"
            request = requests.Request("PATCH", url)
            if self._change_vector is not None:
                request.headers = {"If-Match": f'"{self._change_vector}"'}

            request.data = {
                "Patch": self._patch.serialize(),
                "PatchIfMissing": self._patch_if_missing.serialize() if self._patch_if_missing else None,
            }

            return request

        def is_read_request(self) -> bool:
            return False

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return

            self.result = PatchResult.from_json(json.loads(response))


class PatchByQueryOperation(IOperation[OperationIdResult]):
    def __init__(self, query_to_update: Union[IndexQuery, str], options: Optional[QueryOperationOptions] = None):
        if query_to_update is None:
            raise ValueError("Invalid query")
        super().__init__()
        if isinstance(query_to_update, str):
            query_to_update = IndexQuery(query=query_to_update)
        self._query_to_update = query_to_update
        if options is None:
            options = QueryOperationOptions()
        self._options = options

    def get_command(
        self, store: "DocumentStore", conventions: "DocumentConventions", cache: Optional[HttpCache] = None
    ) -> RavenCommand[OperationIdResult]:
        return self.__PatchByQueryCommand(conventions, self._query_to_update, self._options)

    class __PatchByQueryCommand(RavenCommand[OperationIdResult]):
        def __init__(
            self, conventions: "DocumentConventions", query_to_update: IndexQuery, options: QueryOperationOptions
        ):
            super().__init__(OperationIdResult)
            self._conventions = conventions
            self._query_to_update = query_to_update
            self._options = options

        def create_request(self, server_node) -> requests.Request:
            if not isinstance(self._query_to_update, IndexQuery):
                raise ValueError("query must be IndexQuery Type")

            url = server_node.url + "/databases/" + server_node.database + "/queries"
            path = (
                f"?allowStale={self._options.allow_stale}&maxOpsPerSec={self._options.max_ops_per_sec or ''}"
                f"&details={self._options.retrieve_details}"
            )
            if self._options.stale_timeout is not None:
                path += "&staleTimeout=" + str(self._options.stale_timeout)

            url += path
            data = {"Query": self._query_to_update.to_json()}
            return requests.Request("PATCH", url, data=data)

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()
            response = json.loads(response)
            self.result = OperationIdResult(response["OperationId"], response["OperationNodeTag"])

        def is_read_request(self) -> bool:
            return False



from __future__ import annotations
import datetime
import http
import json
from typing import Optional, List
import requests

from ravendb.documents.session.time_series import (
    TimeSeriesRange,
    TimeSeriesTimeRange,
    TimeSeriesCountRange,
    AbstractTimeSeriesRange,
)
from ravendb.primitives import constants
from ravendb.documents.commands.results import GetDocumentsResult
from ravendb.documents.queries.utils import HashCalculator
from ravendb.http.misc import ResponseDisposeHandling
from ravendb.http.raven_command import VoidRavenCommand
from ravendb.documents.identity.hilo import HiLoResult
from ravendb.extensions.http_extensions import HttpExtensions
from ravendb.http.http_cache import HttpCache
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.tools.utils import Utils


class PutResult:
    def __init__(self, key: Optional[str] = None, change_vector: Optional[str] = None):
        self.key = key
        self.change_vector = change_vector

    @classmethod
    def from_json(cls, json_dict: dict) -> PutResult:
        return cls(json_dict["Id"], json_dict["ChangeVector"])


class PutDocumentCommand(RavenCommand[PutResult]):
    def __init__(self, key: str, change_vector: str, document: dict):
        if not key:
            raise ValueError("Key cannot be None")

        if document is None:
            raise ValueError("Document cannot be None")

        super().__init__(PutResult)
        self.__key = key
        self.__change_vector = change_vector
        self.__document = document

    def create_request(self, node: ServerNode) -> requests.Request:
        request = requests.Request("PUT", f"{node.url}/databases/{node.database}/docs?id={self.__key}")
        request.data = json.loads(json.dumps(self.__document))
        self._add_change_vector_if_not_none(self.__change_vector, request)
        return request

    def set_response(self, response: str, from_cache: bool) -> None:
        self.result = PutResult.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return False


class DeleteDocumentCommand(VoidRavenCommand):
    def __init__(self, key: str, change_vector: Optional[str] = None):
        if not key:
            raise ValueError("Key cannot be None")
        super().__init__()
        self.__key = key
        self.__change_vector = change_vector

    def create_request(self, node: ServerNode) -> requests.Request:
        self.ensure_is_not_null_or_string(self.__key, "id")
        request = requests.Request(
            "DELETE", f"{node.url}/databases/{node.database}/docs?id={Utils.escape(self.__key, None, None)}"
        )
        self._add_change_vector_if_not_none(self.__change_vector, request)
        return request


class HeadDocumentCommand(RavenCommand[str]):
    def __init__(self, key: str, change_vector: Optional[str]):
        super(HeadDocumentCommand, self).__init__(str)
        if key is None:
            raise ValueError("Key cannot be None")
        self.__key = key
        self.__change_vector = change_vector

    def is_read_request(self) -> bool:
        return False

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/databases/{node.database}/docs?id={Utils.escape(self.__key, True, False)}"
        request = requests.Request("HEAD", url)
        if self.__change_vector is not None:
            request.headers[constants.Headers.IF_NONE_MATCH] = self.__change_vector
        return request

    def process_response(self, cache: HttpCache, response: requests.Response, url) -> ResponseDisposeHandling:
        if http.HTTPStatus.NOT_MODIFIED == response.status_code:
            self.result = self.__change_vector
            return ResponseDisposeHandling.AUTOMATIC

        if response.status_code == http.HTTPStatus.NOT_FOUND:
            self.result = None
            return ResponseDisposeHandling.AUTOMATIC

        self.result = HttpExtensions.get_required_etag_header(response)
        return ResponseDisposeHandling.AUTOMATIC

    def set_response(self, response: str, from_cache: bool) -> None:
        if response is not None:
            raise ValueError("Invalid response")

        self.result = None


class GetDocumentsCommand(RavenCommand[GetDocumentsResult]):
    def __init__(self):
        """Creates hollow GetDocumentsCommand.
        Class methods ('from_*') instantiate this class by this constructor - then the object is filled with data."""
        super(GetDocumentsCommand, self).__init__(GetDocumentsResult)
        self._keys: Optional[List[str]] = None
        self._includes: Optional[List[str]] = None
        self._metadata_only: Optional[bool] = None

        self._key: Optional[str] = None
        self._counters: Optional[List[str]] = None
        self._include_all_counters: Optional[bool] = None
        self._time_series_includes: Optional[List[AbstractTimeSeriesRange]] = None

        self._compare_exchange_value_includes: Optional[List[str]] = None

        self._start_with: Optional[str] = None
        self._matches: Optional[str] = None
        self._start: Optional[int] = None
        self._page_size: Optional[int] = None
        self._exclude: Optional[str] = None
        self._start_after: Optional[str] = None

    @classmethod
    def from_paging(cls, start: int, page_size: int) -> GetDocumentsCommand:
        command = cls()
        command._start = start
        command._page_size = page_size
        return command

    @classmethod
    def from_single_id(cls, key: str, includes: List[str] = None, metadata_only: bool = None) -> GetDocumentsCommand:
        if key is None:
            raise ValueError("Key cannot be None")
        command = cls()
        command._key = key
        command._includes = includes
        command._metadata_only = metadata_only
        return command

    @classmethod
    def from_multiple_ids(
        cls,
        keys: List[str],
        includes: List[str] = None,
        counter_includes: List[str] = None,
        time_series_includes: List[str] = None,
        compare_exchange_value_includes: List[str] = None,
        metadata_only: bool = False,
    ) -> GetDocumentsCommand:
        if not keys:
            raise ValueError("Please supply at least one id")
        command = cls()
        command._keys = keys
        command._includes = includes
        command._metadata_only = metadata_only

        command._counters = counter_includes
        command._time_series_includes = time_series_includes
        command._compare_exchange_value_includes = compare_exchange_value_includes
        return command

    @classmethod
    def from_multiple_ids_all_counters(
        cls,
        keys: List[str],
        includes: List[str] = None,
        include_all_counters: bool = None,
        time_series_includes: List[str] = None,
        compare_exchange_value_includes: List[str] = None,
        metadata_only: bool = False,
    ) -> GetDocumentsCommand:
        if not keys:
            raise ValueError("Please supply at least one id")
        command = cls()
        command._keys = keys
        command._includes = includes
        command._include_all_counters = include_all_counters
        command._time_series_includes = time_series_includes
        command._compare_exchange_value_includes = compare_exchange_value_includes
        command._metadata_only = metadata_only
        return command

    @classmethod
    def from_starts_with(
        cls,
        start_with: str,
        start_after: str = None,
        matches: str = None,
        exclude: str = None,
        start: int = None,
        page_size: int = None,
        metadata_only: bool = None,
    ) -> GetDocumentsCommand:
        if start_with is None:
            raise ValueError("start_with cannot be None")

        command = cls()
        command._start_with = start_with
        command._start_after = start_after
        command._matches = matches
        command._exclude = exclude
        command._start = start
        command._page_size = page_size
        command._metadata_only = metadata_only

        return command

    def create_request(self, node: ServerNode) -> requests.Request:
        path_builder: List[str] = [node.url, f"/databases/{node.database}/docs?"]

        if self._start is not None:
            path_builder.append(f"&start={self._start}")

        if self._page_size is not None:
            path_builder.append(f"&pageSize={self._page_size}")

        if self._metadata_only:
            path_builder.append("&metadataOnly=true")

        if self._start_with is not None:
            path_builder.append(f"&startsWith={Utils.quote_key(self._start_with)}")

            if self._matches is not None:
                path_builder.append(f"&matches={self._matches}")

            if self._exclude is not None:
                path_builder.append(f"&exclude={self._exclude}")

            if self._start_after is not None:
                path_builder.append(f"&startAfter={self._start_after}")

        if self._includes is not None:
            for include in self._includes:
                path_builder.append(f"&include={include}")

        if self._include_all_counters:
            path_builder.append(f"&counter={constants.Counters.ALL}")
        elif self._counters:
            for counter in self._counters:
                path_builder.append(f"&counter={counter}")

        if self._time_series_includes is not None:
            for ts_include in self._time_series_includes:
                if isinstance(ts_include, TimeSeriesRange):
                    range_: TimeSeriesRange = ts_include
                    path_builder.append(
                        f"&timeseries={range_.name}"
                        f"&from={Utils.datetime_to_string(range_.from_date) if range_.from_date else ''}"
                        f"&to={Utils.datetime_to_string(range_.to_date) if range_.to_date else ''}"
                    )
                elif isinstance(ts_include, TimeSeriesTimeRange):
                    time_range: TimeSeriesTimeRange = ts_include
                    path_builder.append(
                        f"&timeseriestime={time_range.name}"
                        f"&timeType={time_range.type.value}"
                        f"&timeValue={time_range.time.value}"
                        f"&timeUnit={time_range.time.unit.value}"
                    )
                elif isinstance(ts_include, TimeSeriesCountRange):
                    count_range: TimeSeriesCountRange = ts_include
                    path_builder.append(
                        f"&timeseriescount={count_range.name}"
                        f"&countType={count_range.type.value}"
                        f"&countValue={count_range.count}"
                    )
                else:
                    raise TypeError(f"Unexpected TimeSeries range {ts_include.__class__.__name__}")

        if self._compare_exchange_value_includes is not None:
            for compare_exchange_value in self._compare_exchange_value_includes:
                path_builder.append(f"&cmpxchg={compare_exchange_value}")

        request = requests.Request("GET")

        if self._key is not None:
            path_builder.append(f"&id={Utils.quote_key(self._key)}")
        if self._keys is not None:
            request = self.prepare_request_with_multiple_ids(path_builder, request, self._keys)

        request.url = "".join(path_builder)
        return request

    def prepare_request_with_multiple_ids(
        self, path_builder: List[str], request: requests.Request, ids: List[str]
    ) -> requests.Request:
        unique_ids = set()
        unique_ids.update(ids)
        is_get = sum(map(len, list(filter(lambda x: x, unique_ids)))) < 1024

        if is_get:
            for key in unique_ids:
                path_builder.append(f"&id={Utils.quote_key(key) if key else ''}")

            return requests.Request("GET")

        else:
            http_post = requests.Request("POST")
            try:
                calculate_hash = HashCalculator.calculate_hash_from_str_collection(unique_ids)
                path_builder.append(f"&loadHash={calculate_hash}")
            except Exception as e:
                raise RuntimeError(f"Unable to computer query hash: {e.args[0]}", e)

            http_post.data = {"Ids": [str(key) for key in unique_ids]}
            return http_post

    def set_response(self, response: str, from_cache: bool) -> None:
        self.result = GetDocumentsResult.from_json(json.loads(response)) if response is not None else None

    @property
    def is_read_request(self) -> bool:
        return True


class NextHiLoCommand(RavenCommand[HiLoResult]):
    def __init__(
        self,
        tag: str,
        last_batch_size: int,
        last_range_at: datetime.datetime,
        identity_parts_separator: str,
        last_range_max: int,
    ):
        super().__init__(HiLoResult)
        self.__tag = tag
        self.__last_batch_size = last_batch_size
        self.__last_range_at = last_range_at
        self.__identity_parts_separator = identity_parts_separator
        self.__last_range_max = last_range_max

    def is_read_request(self) -> bool:
        return True

    def create_request(self, node: ServerNode) -> requests.Request:
        date = Utils.datetime_to_string(self.__last_range_at) if self.__last_range_at else ""
        path = (
            f"/hilo/next?tag={Utils.escape(self.__tag, False, False)}"
            f"&lastBatchSize={self.__last_batch_size or 0}"
            f"&lastRangeAt={date}"
            f"&identityPartsSeparator={self.__identity_parts_separator}"
            f"&lastMax={self.__last_range_max}"
        )

        url = f"{node.url}/databases/{node.database}{path}"
        return requests.Request(method="GET", url=url)

    def set_response(self, response: str, from_cache: bool) -> None:
        self.result = HiLoResult.from_json(json.loads(response))


class HiLoReturnCommand(VoidRavenCommand):
    def __init__(self, tag: str, last: int, end: int):
        super().__init__()
        if last < 0:
            raise ValueError("Last is < 0")

        if end < 0:
            raise ValueError("End is <0")

        if tag is None:
            raise ValueError("tag cannot be None")

        self.__tag = tag
        self.__last = last
        self.__end = end

    def create_request(self, node: ServerNode) -> requests.Request:
        return requests.Request(
            "PUT",
            url=f"{node.url}/databases/{node.database}/hilo/return?tag={self.__tag}&end={self.__end}&last={self.__last}",
        )


class HeadAttachmentCommand(RavenCommand[str]):
    def __init__(self, document_id: str, name: str, change_vector: Optional[str]):
        super().__init__(str)

        if document_id.isspace():
            raise ValueError("Document_id cannot be None or whitespace")

        if name.isspace():
            raise ValueError("Name cannot be None or whitespace")

        self.__document_id = document_id
        self.__name = name
        self.__change_vector = change_vector

    def create_request(self, node: ServerNode) -> requests.Request:
        return requests.Request(
            "HEAD",
            f"{node.url}/databases/{node.database}"
            f"/attachments?id={Utils.quote_key(self.__document_id)}"
            f"&name={Utils.quote_key(self.__name)}",
            {constants.Headers.IF_NONE_MATCH: self.__change_vector} if self.__change_vector else None,
        )

    def process_response(self, cache: HttpCache, response: requests.Response, url) -> http.ResponseDisposeHandling:
        if response.status_code == http.HTTPStatus.NOT_MODIFIED:
            self.result = self.__change_vector
            return ResponseDisposeHandling.AUTOMATIC

        if response.status_code == http.HTTPStatus.NOT_FOUND:
            self.result = None
            return ResponseDisposeHandling.AUTOMATIC

        chv = response.headers.get(constants.Headers.ETAG, None)
        if not chv:
            raise RuntimeError("Response didn't had an ETag header")
        self.result = chv
        return ResponseDisposeHandling.AUTOMATIC

    def set_response(self, response: str, from_cache: bool) -> None:
        if response is not None:
            self._throw_invalid_response()

        self.result = None

    def is_read_request(self) -> bool:
        return False


class ConditionalGetResult:
    def __init__(self, results: List = None, change_vector: str = None):
        self.results = results
        self.change_vector = change_vector

    @classmethod
    def from_json(cls, json_dict: dict) -> ConditionalGetResult:
        return cls(json_dict["Results"], None)


class ConditionalGetDocumentsCommand(RavenCommand[ConditionalGetResult]):
    def __init__(self, key: str, change_vector: str):
        super().__init__(ConditionalGetResult)

        self.__change_vector = change_vector
        self.__key = key

    def create_request(self, node: ServerNode) -> requests.Request:
        url = f"{node.url}/databases/{node.database}/docs?id={Utils.quote_key(self.__key)}"
        request = requests.Request("GET", url)
        request.headers[constants.Headers.IF_NONE_MATCH] = f'"{self.__change_vector}"'

        return request

    def set_response(self, response: str, from_cache: bool) -> None:
        if response is None:
            self.result = None
            return

        self.result = ConditionalGetResult.from_json(json.loads(response))

    def is_read_request(self) -> bool:
        return False

    def process_response(self, cache: HttpCache, response: requests.Response, url) -> ResponseDisposeHandling:
        if response.status_code == http.HTTPStatus.NOT_MODIFIED:
            return ResponseDisposeHandling.AUTOMATIC

        result = super().process_response(cache, response, url)
        self.result.change_vector = response.headers["ETag"]
        return result



from __future__ import annotations
from enum import Enum
from typing import List, Tuple, Dict, Any

from ravendb.primitives.constants import int_max, int_min


class TimeValueUnit(Enum):
    NONE = "None"
    SECOND = "Second"
    MONTH = "Month"


class TimeValue:
    SECONDS_PER_DAY = 86400
    SECONDS_IN_28_DAYS = 28 * SECONDS_PER_DAY  # lower-bound of seconds in month
    SECONDS_IN_31_DAYS = 31 * SECONDS_PER_DAY  # upper-bound of seconds in month
    SECONDS_IN_365_DAYS = 365 * SECONDS_PER_DAY  # lower-bound of seconds in year
    SECONDS_IN_366_DAYS = 366 * SECONDS_PER_DAY  # upper-bound of seconds in year

    def __init__(self, value: int, unit: TimeValueUnit):
        self.value = value
        self.unit = unit

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> TimeValue:
        return cls(json_dict["Value"], TimeValueUnit(json_dict["Unit"]))

    def to_json(self) -> Dict[str, Any]:
        return {"Value": self.value, "Unit": self.unit.value}

    def __str__(self):
        if self.value == int_max:
            return "MaxValue"
        if self.value == int_min:
            return "MinValue"
        if self.value == 0:
            return "Zero"

        if self.unit == TimeValueUnit.NONE:
            return "Unknown time unit"

        builder = []

        if self.unit == TimeValueUnit.SECOND:
            remaining_seconds = self.value

            if remaining_seconds > self.SECONDS_PER_DAY:
                days = self.value // self.SECONDS_PER_DAY
                self._append(builder, days, "day")
                remaining_seconds -= days * self.SECONDS_PER_DAY

            if remaining_seconds > 3600:
                hours = remaining_seconds // 3600
                self._append(builder, hours, "hour")
                remaining_seconds -= hours * 3600

            if remaining_seconds > 60:
                minutes = remaining_seconds // 60
                self._append(builder, minutes, "minute")
                remaining_seconds -= minutes * 60

            if remaining_seconds > 0:
                self._append(builder, remaining_seconds, "second")

        elif self.unit == TimeValueUnit.MONTH:
            if self.value >= 12:
                self._append(builder, self.value // 12, "year")
            if self.value % 12 > 0:
                self._append(builder, self.value % 12, "month")

        else:
            raise ValueError(f"Not supported unit: {self.unit}")

        return "".join(builder).strip()

    @classmethod
    def ZERO(cls):
        return cls(0, TimeValueUnit.NONE)

    @classmethod
    def MAX_VALUE(cls):
        return cls(int_max, TimeValueUnit.NONE)

    @classmethod
    def MIN_VALUE(cls):
        return cls(int_min, TimeValueUnit.NONE)

    @classmethod
    def of_seconds(cls, seconds: int) -> TimeValue:
        return cls(seconds, TimeValueUnit.SECOND)

    @classmethod
    def of_minutes(cls, minutes: int) -> TimeValue:
        return cls(minutes * 60, TimeValueUnit.SECOND)

    @classmethod
    def of_hours(cls, hours: int) -> TimeValue:
        return cls(hours * 3600, TimeValueUnit.SECOND)

    @classmethod
    def of_days(cls, days: int) -> TimeValue:
        return cls(days * cls.SECONDS_PER_DAY, TimeValueUnit.SECOND)

    @classmethod
    def of_months(cls, months: int) -> TimeValue:
        return cls(months, TimeValueUnit.MONTH)

    @classmethod
    def of_years(cls, years: int) -> TimeValue:
        return cls(12 * years, TimeValueUnit.MONTH)

    def _append(self, builder: List[str], value: int, singular: str) -> None:
        if value <= 0:
            return

        builder.append(str(value))
        builder.append(" ")
        builder.append(singular)

        if value == 1:
            builder.append(" ")
            return

        builder.append("s ")  # lucky me, no special rules here

    def _assert_seconds(self) -> None:
        if self.unit != TimeValueUnit.SECOND:
            raise ValueError("The value must be seconds")

    @staticmethod
    def _assert_valid_unit(unit: TimeValueUnit) -> None:
        if unit == TimeValueUnit.MONTH or unit == TimeValueUnit.SECOND:
            return

        raise ValueError(f"Invalid time unit: {unit}")

    @staticmethod
    def _assert_same_units(a: TimeValue, b: TimeValue) -> None:
        if a.unit != b.unit:
            raise ValueError(f"Unit isn't the same {a.unit} != {b.unit}")

    @classmethod
    def _get_bounds_in_seconds(cls, time: TimeValue) -> Tuple[int, int]:
        if time.unit == TimeValueUnit.SECOND:
            return time.value, time.value
        elif time.unit == TimeValueUnit.MONTH:
            years = time.value // 12
            upper_bound = years * cls.SECONDS_IN_366_DAYS
            lower_bound = years * cls.SECONDS_IN_365_DAYS

            remaining_months = time.value % 12
            upper_bound += remaining_months * cls.SECONDS_IN_31_DAYS
            lower_bound += remaining_months * cls.SECONDS_IN_28_DAYS

            return lower_bound, upper_bound
        else:
            raise ValueError(f"Not supported time value unit: {time.unit}")

    def compare_to(self, other: TimeValue) -> int:
        if self.value == 0 or other.value == 0:
            return self.value - other.value

        condition, result = self.is_special_compare(self, other)
        if condition:
            return result

        if self.unit == other.unit:
            return self._trim_compare_result(self.value - other.value)

        my_bounds = self._get_bounds_in_seconds(self)
        other_bounds = self._get_bounds_in_seconds(other)

        if other_bounds[1] < my_bounds[0]:
            return 1

        if other_bounds[0] > my_bounds[1]:
            return -1

        raise ValueError(f"Unable to compare {self} with {other}, since a month might have different number of days.")

    @staticmethod
    def _is_max(time: TimeValue) -> bool:
        return time.unit == TimeValueUnit.NONE and time.value == int_max

    @staticmethod
    def _is_min(time: TimeValue) -> bool:
        return time.unit == TimeValueUnit.NONE and time.value == int_min

    @staticmethod
    def _trim_compare_result(result: int) -> int:
        if result > int_max:
            return int_max

        if result < int_min:
            return int_min

        return result

    def __eq__(self, o: object):
        if id(self) == id(o):
            return True
        if o is None or self.__class__ != o.__class__:
            return False
        other: TimeValue = o
        return self.compare_to(other) == 0

    def __hash__(self):
        if self.value == 0 or self.value == int_min or self.value == int_max:
            return hash(self.value)
        return hash((self.value, self.unit))

    @classmethod
    def is_special_compare(cls, current: TimeValue, other: TimeValue) -> Tuple[bool, int]:
        result = 0

        if cls._is_max(current):
            result = 0 if cls._is_max(other) else 1
            return True, result

        if cls._is_max(other):
            result = 0 if cls._is_max(current) else -1
            return True, result

        if cls._is_min(current):
            result = 0 if cls._is_min(other) else -1
            return True, result

        if cls._is_min(other):
            result = 0 if cls._is_min(current) else 1
            return True, result

        return False, result



import base64
import ssl
from threading import Lock
from typing import TYPE_CHECKING, Dict, Optional, Callable, Any, List

from websocket import WebSocket

from ravendb.changes.observers import Observable
from ravendb.changes.types import (
    DocumentChange,
    IndexChange,
    TimeSeriesChange,
    CounterChange,
    OperationStatusChange,
    TopologyChange,
    DatabaseChange,
)
from ravendb.serverwide.commands import GetTcpInfoCommand
from ravendb.tools.parsers import IncrementalJsonParser
import websocket
from ravendb.exceptions.exceptions import NotSupportedException
from ravendb.exceptions.exceptions import ChangeProcessingException
from ravendb.tools.utils import Utils
import copy
from time import sleep
from concurrent.futures import ThreadPoolExecutor, TimeoutError, Future
import logging
import sys

if TYPE_CHECKING:
    from ravendb.http.request_executor import RequestExecutor


class DatabaseChanges:
    def __init__(
        self,
        request_executor: "RequestExecutor",
        database_name: str,
        on_close: Callable[[str], None],
        on_error: Optional[Callable[[Exception], None]] = None,
        executor: Optional[ThreadPoolExecutor] = None,
    ):
        self._request_executor = request_executor
        self._conventions = request_executor.conventions
        self._database_name = database_name

        self._command_id = 0
        self.client_websocket = websocket.WebSocket()
        self._closed = False
        self._on_close = on_close
        self.on_error = on_error
        self._observables_by_group: Dict[str, Dict[str, Observable[DatabaseChange]]] = {}

        self._executor = executor if executor else ThreadPoolExecutor(max_workers=10)
        self._worker = self._executor.submit(self.do_work)
        self.send_lock = Lock()
        self._confirmations_lock = Lock()
        self._confirmations: Dict[int, Future] = {}

        self._command_id = 0
        self._immediate_connection = 0

        self._logger = logging.getLogger("database_changes")
        handler = logging.FileHandler("changes.log")
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        self._logger.addHandler(handler)
        self._logger.setLevel(logging.DEBUG)

    def _ensure_websocket_connected(self, url: str) -> None:
        if self._request_executor.certificate_path:
            self._connect_websocket_secured(url)
        else:
            self.client_websocket.connect(url)

        for observables_by_name in self._observables_by_group.values():
            for observer in observables_by_name.values():
                observer.set(self._executor.submit(observer.on_connect))
        self._immediate_connection = 1

    def _get_server_certificate(self) -> Optional[str]:
        cmd = GetTcpInfoCommand(self._request_executor.url)
        self._request_executor.execute_command(cmd)
        return cmd.result.certificate

    def _connect_websocket_secured(self, url: str) -> None:
        # Get server certificate via HTTPS and prepare SSL context
        server_certificate = base64.b64decode(self._get_server_certificate())
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        ssl_context.load_cert_chain(self._request_executor.certificate_path)
        if self._request_executor.trust_store_path:
            ssl_context.verify_mode = ssl.CERT_REQUIRED
            ssl_context.load_verify_locations(self._request_executor.trust_store_path)

        # Create SSL WebSocket and connect it
        self.client_websocket = WebSocket(sslopt={"context": ssl_context})
        self.client_websocket.connect(url, suppress_origin=True)

        # Server certificate authentication
        server_certificate_from_tls = self.client_websocket.sock.getpeercert(True)
        if server_certificate != server_certificate_from_tls:
            raise ValueError("Certificates don't match")

    def do_work(self):
        preferred_node = self._request_executor.preferred_node.current_node
        url = (
            f"{preferred_node.url}/databases/{self._database_name}/changes".replace("http://", "ws://")
            .lower()
            .replace("https://", "wss://")
            .replace(".fiddler", "")
        )

        while not self._closed:
            try:
                if not self.client_websocket.connected:
                    self._ensure_websocket_connected(url)
                self.process_changes()
            except ChangeProcessingException as e:
                self.notify_about_error(e)
                continue
            except websocket.WebSocketException as e:
                if getattr(sys, "gettrace", None):
                    self._logger.error(e)
                self.notify_about_error(e)
                return
            except Exception as e:
                self._immediate_connection = 0
                self.client_websocket = websocket.WebSocket()
                self.notify_about_error(e)
            sleep(1)

    def process_changes(self):
        parser = IncrementalJsonParser(self.client_websocket, is_websocket=True)
        while not self._closed:
            try:
                response = parser.next_object()
                if not response:
                    continue

                response_type: Optional[str] = response.get("Type", None)
                if not response_type:
                    continue

                if response_type == "Error":
                    exception = response["Exception"]
                    self.notify_about_error(Exception(exception))
                elif response_type == "Confirm":
                    command_id: Optional[int] = response.get("CommandId", None)
                    if not command_id or command_id not in self._confirmations:
                        continue
                    with self._confirmations_lock:
                        future = self._confirmations.pop(command_id)
                        future.set_result("done complete future")
                else:
                    change_json_dict: Optional[Dict[str, Any]] = response.get("Value", None)
                    self._notify_subscribers(
                        response_type, change_json_dict, copy.copy(self._observables_by_group[response_type])
                    )
            except Exception as e:
                self.notify_about_error(e)
                raise ChangeProcessingException(e)

    @staticmethod
    def _notify_subscribers(type_of_change: str, change_json_dict: Dict[str, Any], observables: Dict[str, Observable]):
        if type_of_change == "DocumentChange":
            result = DocumentChange.from_json(change_json_dict)
        elif type_of_change == "IndexChange":
            result = IndexChange.from_json(change_json_dict)
        elif type_of_change == "TimeSeriesChange":
            result = TimeSeriesChange.from_json(change_json_dict)
        elif type_of_change == "CounterChange":
            result = CounterChange.from_json(change_json_dict)
        elif type_of_change == "OperationStatusChange":
            result = OperationStatusChange.from_json(change_json_dict)
        elif type_of_change == "TopologyChange":
            result = TopologyChange.from_json(change_json_dict)
        else:
            raise NotSupportedException(type_of_change)

        for observable in observables.values():
            observable.send(result)

    def close(self):
        self._closed = True
        self.client_websocket.close()

        for observable in self._observables_by_group.values():
            for observer in observable.values():
                observer.close()

        with self._confirmations_lock:
            for confirmation in self._confirmations.values():
                confirmation.cancel()

        self._observables_by_group.clear()
        if self._on_close:
            self._on_close(self._database_name)

        self._executor.shutdown(wait=True)

    def notify_about_error(self, e: Exception):
        if self.on_error:
            self.on_error(e)

        for _, observables in self._observables_by_group.items():
            for observer in observables.values():
                observer.error(e)

    def for_all_documents(self) -> Observable[DocumentChange]:
        observable = self.get_or_add_observable("DocumentChange", "all-docs", "watch-docs", "unwatch-docs")(
            lambda x: True
        )
        return observable

    def for_all_operations(self) -> Observable[OperationStatusChange]:
        observable = self.get_or_add_observable(
            "OperationsStatusChange",
            "all-operations",
            "watch-operations",
            "unwatch-operations",
        )(lambda x: True)
        return observable

    def for_all_indexes(self) -> Observable[IndexChange]:
        observable = self.get_or_add_observable("IndexChange", "all-indexes", "watch-indexes", "unwatch-indexes")(
            lambda x: True
        )
        return observable

    def for_index(self, index_name: str) -> Observable[IndexChange]:
        observable = self.get_or_add_observable(
            "IndexChange",
            "indexes/" + index_name,
            "watch-index",
            "unwatch-index",
            index_name,
        )(lambda x: x.name.casefold() == index_name.casefold())
        return observable

    def for_operation_id(self, operation_id: int) -> Observable[OperationStatusChange]:
        observable = self.get_or_add_observable(
            "OperationsStatusChange",
            "operations/" + str(operation_id),
            "watch-operation",
            "unwatch-operation",
            str(operation_id),
        )(lambda x: x.operation_id == str(operation_id))
        return observable

    def for_document(self, doc_id: str) -> Observable[DocumentChange]:
        observable = self.get_or_add_observable("DocumentChange", "docs/" + doc_id, "watch-doc", "unwatch-doc", doc_id)(
            lambda x: x.key.casefold() == doc_id.casefold()
        )
        return observable

    def for_documents_start_with(self, doc_id_prefix: str) -> Observable[DocumentChange]:
        observable = self.get_or_add_observable(
            "DocumentChange",
            "prefixes/" + doc_id_prefix,
            "watch-prefix",
            "unwatch-prefix",
            doc_id_prefix,
        )(lambda x: x.key is not None and x.key.casefold().startswith(doc_id_prefix.casefold()))
        return observable

    def for_documents_in_collection(self, collection_name: str) -> Observable[DocumentChange]:
        observable = self.get_or_add_observable(
            "DocumentChange",
            "collections/" + collection_name,
            "watch-collection",
            "unwatch-collection",
            collection_name,
        )(lambda x: x.collection_name.casefold() == collection_name.casefold())
        return observable

    def for_all_time_series(self) -> Observable[TimeSeriesChange]:
        observable = self.get_or_add_observable(
            "TimeSeriesChange",
            "all-timeseries",
            "watch-all-timeseries",
            "unwatch-all-timeseries",
        )(lambda x: True)
        return observable

    def for_time_series(self, time_series_name: str) -> Observable[TimeSeriesChange]:
        if not time_series_name:
            raise ValueError("time_series_name cannot be None or empty")
        observable = self.get_or_add_observable(
            "TimeSeriesChange",
            f"timeseries/{time_series_name}",
            "watch-timeseries",
            "unwatch-timeseries",
            time_series_name,
        )(lambda x: x.name.casefold() == time_series_name.casefold())
        return observable

    def for_time_series_of_document(self, doc_id: str, time_series_name: str = None) -> Observable[TimeSeriesChange]:
        """
        Can subscribe to all time series changes that associated with the document or
        by passing the time series name only for a specific time series
        """
        if not doc_id:
            raise ValueError("doc_id cannot be None or empty")

        def get_lambda():
            if time_series_name:
                return lambda x: x.document_id.casefold() == doc_id.casefold() and x.name.casefold()
            return lambda x: x.document_id.casefold() == doc_id.casefold()

        name = f"document/{doc_id}/timeseries{f'/{time_series_name}' if time_series_name else ''}"
        watch_command = "watch-document-timeseries" if time_series_name else "watch-all-document-timeseries"
        unwatch_command = "unwatch-document-timeseries" if time_series_name else "unwatch-all-document-timeseries"
        value = doc_id if time_series_name is None else None
        values = [doc_id, time_series_name] if time_series_name is not None else None
        observable = self.get_or_add_observable(
            "TimeSeriesChange",
            name,
            watch_command,
            unwatch_command,
            resource_name=value,
            resources_names=values,
        )(get_lambda())
        return observable

    def for_all_counters(self) -> Observable[CounterChange]:
        observable = self.get_or_add_observable("CounterChange", "all-counters", "watch-counters", "unwatch-counters")(
            lambda x: True
        )
        return observable

    def for_counter(self, counter_name: str) -> Observable[CounterChange]:
        if not counter_name:
            raise ValueError("counter_name cannot be None or empty")
        observable = self.get_or_add_observable(
            "CounterChange",
            f"counter/{counter_name}",
            "watch-counter",
            "unwatch-counter",
            counter_name,
        )(lambda x: x.name.casefold() == counter_name.casefold())
        return observable

    def for_counters_of_document(self, doc_id: str) -> Observable[CounterChange]:
        """
        Can subscribe to all counters changes that associated with the document or
        """
        if not doc_id:
            raise ValueError("doc_id cannot be None or empty")

        observable = self.get_or_add_observable(
            "CounterChange",
            f"document/{doc_id}/counter",
            "watch-document-counters",
            "unwatch-document-counters",
            resource_name=doc_id,
        )(lambda x: x.document_id.casefold() == doc_id.casefold())
        return observable

    def for_counter_of_document(self, doc_id: str, counter_name: str) -> Observable[CounterChange]:
        """
        Can subscribe to all counters changes that associated with the document and for counter name
        """
        if not doc_id:
            raise ValueError("doc_id cannot be None or empty")
        if not counter_name:
            raise ValueError("counter_name cannot be None or empty")

        observable = self.get_or_add_observable(
            "CounterChange",
            f"document/{doc_id}/counter/{counter_name}",
            "watch-document-counter",
            "unwatch-document-counter",
            resources_names=[doc_id, counter_name],
        )(lambda x: x.document_id.casefold() == doc_id.casefold() and x.name.casefold())
        return observable

    def get_or_add_observable(
        self,
        group: str,
        name: str,
        watch_command: str,
        unwatch_command: str,
        resource_name: Optional[str] = None,
        resources_names: Optional[List[str]] = None,
    ):
        if group not in self._observables_by_group:
            self._observables_by_group[group] = {}

        if name not in self._observables_by_group[group]:

            def on_disconnect():
                try:
                    if self.client_websocket.connected:
                        self.send(unwatch_command, resource_name, resources_names)
                except websocket.WebSocketException:
                    pass

            def on_connect():
                self.send(watch_command, resource_name, resources_names)

            observable = Observable(
                on_connect=on_connect,
                on_disconnect=on_disconnect,
                executor=self._executor,
            )
            self._observables_by_group[group][name] = observable
            if self._immediate_connection != 0:
                observable.set(self._executor.submit(observable.on_connect))
        return self._observables_by_group[group][name]

    def send(self, command: str, value: Optional[str], values: Optional[List[str]] = None):
        current_command_id = 0
        future = Future()
        try:
            with self.send_lock:
                self._command_id += 1
                current_command_id += self._command_id
                data_dict = {
                    "CommandId": current_command_id,
                    "Command": command,
                    "Param": value,
                }
                if values:
                    data_dict["Params"] = values
                data = Utils.dict_to_bytes(data_dict)
                with self._confirmations_lock:
                    self._confirmations[current_command_id] = future
                self.client_websocket.send(data)

                try:
                    future.result(timeout=15)
                except TimeoutError:
                    future.cancel()
                    raise TimeoutError("Did not get a confirmation for command #" + str(current_command_id))
                except Exception as e:
                    if getattr(sys, "gettrace", None):
                        self._logger.info("The coroutine raised an exception: {!r}".format(e))
        except websocket.WebSocketConnectionClosedException:
            pass



from typing import List, TYPE_CHECKING

import requests

from ravendb import RaftCommand, ServerNode
from ravendb.documents.indexes.analysis.definitions import AnalyzerDefinition
from ravendb.http.raven_command import VoidRavenCommand
from ravendb.serverwide.operations.common import VoidServerOperation
from ravendb.tools.utils import Utils
from ravendb.util.util import RaftIdGenerator

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class PutServerWideAnalyzersOperation(VoidServerOperation):
    def __init__(self, *analyzers_to_add: AnalyzerDefinition):
        if not analyzers_to_add:
            raise ValueError("Analyzers must not be empty or None")

        self._analyzers_to_add = analyzers_to_add

    def get_command(
        self, conventions: "DocumentConventions"
    ) -> "PutServerWideAnalyzersOperation.PutServerWideAnalyzersCommand":
        return self.PutServerWideAnalyzersCommand(conventions, list(self._analyzers_to_add))

    class PutServerWideAnalyzersCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, conventions: "DocumentConventions", analyzers_to_add: List[AnalyzerDefinition]):
            if not conventions:
                raise ValueError("Conventions cannot be None")

            if analyzers_to_add is None:
                raise ValueError("Analyzers cannot be None")

            super().__init__()

            self._analyzers_to_add = []
            for i in range(len(analyzers_to_add)):
                analyzer = analyzers_to_add[i]
                if analyzer.name is None:
                    raise ValueError("Name cannot be None")

                self._analyzers_to_add.append(analyzer.to_json())

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/analyzers"

            request = requests.Request(method="PUT", url=url)
            request.data = {"Analyzers": self._analyzers_to_add}

            return request

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class DeleteServerWideAnalyzerOperation(VoidServerOperation):
    def __init__(self, analyzer_name: str):
        if analyzer_name is None:
            raise ValueError("Analyzer name cannot be None")

        self._analyzer_name = analyzer_name

    def get_command(
        self, conventions: "DocumentConventions"
    ) -> "DeleteServerWideAnalyzerOperation.DeleteServerWideAnalyzerCommand":
        return self.DeleteServerWideAnalyzerCommand(self._analyzer_name)

    class DeleteServerWideAnalyzerCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, analyzer_name: str):
            if analyzer_name is None:
                raise ValueError("Analyzer name cannot be None")
            super().__init__()

            self._analyzer_name = analyzer_name

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("DELETE", f"{node.url}/admin/analyzers?name={Utils.quote_key(self._analyzer_name)}")

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



import datetime
from enum import Enum
from typing import Union, Optional, Dict, Set


class ItemFlags(Enum):
    NONE = "None"
    NOT_FOUND = "NotFound"
    AGGRESSIVELY_CACHED = "AggressivelyCached"

    def __str__(self):
        return self.value


class HttpCacheItem:
    def __init__(self):
        self.change_vector: Union[None, str] = None
        self.payload: Union[None, str] = None
        self.last_server_update: datetime.datetime = datetime.datetime.now()
        self.flags: Set[ItemFlags] = {ItemFlags.NONE}
        self.generation: Union[None, int] = None

        self.cache: Union[None, HttpCache] = None


class ReleaseCacheItem:
    def __init__(self, item: Optional[HttpCacheItem] = None):
        self.item = item
        self.__cache_generation = item.cache.generation if item else 0

    def not_modified(self) -> None:
        if self.item is not None:
            self.item.last_server_update = datetime.datetime.now()
            self.item.generation = self.__cache_generation

    @property
    def age(self) -> datetime.timedelta:
        if self.item is None:
            return datetime.timedelta.max
        return datetime.datetime.now() - self.item.last_server_update

    @property
    def might_have_been_modified(self) -> bool:
        return self.item.generation != self.__cache_generation

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class HttpCache:
    NOT_FOUND_RESPONSE = "404 Response"

    def __init__(self):
        self.__items: Dict[str, HttpCacheItem] = {}
        self.generation = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __len__(self):
        return len(self.__items)

    def __setitem__(self, key, value):
        self.__items.__setitem__(key, value)

    def __getitem__(self, item):
        self.__items.__getitem__(item)

    def close(self):
        self.__items.clear()
        self.__items = None

    def clear(self) -> None:
        self.__items.clear()

    def set(self, url: str, change_vector: str, result: str) -> None:
        http_cache_item = HttpCacheItem()
        http_cache_item.change_vector = change_vector
        http_cache_item.payload = result
        http_cache_item.cache = self
        http_cache_item.generation = self.generation
        self.__items[url] = http_cache_item

    def get(self, url: str) -> (ReleaseCacheItem, str, str):
        item = self.__items.get(url, None) if self.__items else None
        if item is not None:
            change_vector = item.change_vector
            response = item.payload
            return ReleaseCacheItem(item), change_vector, response

        return ReleaseCacheItem(), None, None

    def set_not_found(self, url: str, aggressively_cached: bool) -> None:
        http_cache_item = HttpCacheItem()
        http_cache_item.change_vector = self.NOT_FOUND_RESPONSE
        http_cache_item.cache = self
        http_cache_item.generation = self.generation
        http_cache_item.flags = (
            {ItemFlags.AGGRESSIVELY_CACHED, ItemFlags.NOT_FOUND} if aggressively_cached else {ItemFlags.NOT_FOUND}
        )
        self.__items[url] = http_cache_item

    class ReleaseCacheItem:
        def __init__(self, item: HttpCacheItem = None):
            self.item: Union[None, HttpCacheItem] = item
            self.__cache_generation = item.cache.generation if item else 0

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def not_modified(self) -> None:
            if self.item is not None:
                self.item.last_server_update = datetime.datetime.now()
                self.item.generation = self.__cache_generation

        @property
        def age(self) -> datetime.timedelta:
            if self.item is None:
                return datetime.timedelta.max
            return datetime.datetime.now() - self.item.last_server_update

        @property
        def might_have_been_modified(self) -> bool:
            return self.item.generation != self.__cache_generation
